"""
Created on 2020

@author: Eduardo Vitral
"""

###############################################################################
#
# November 2020, Paris
#
# This file contains the main functions concerning positional information.
# It provides MCMC and maximum likelihood fits of surface density,
# as well as robust initial guesses for the (RA,Dec) center of the source.
#
# Documentation is provided on Vitral, 2021.
# If you have any further questions please email vitral@iap.fr
#
###############################################################################

from . import angle
import numpy as np
from skimage.feature import peak_local_max
from scipy.optimize import differential_evolution
from scipy.special import gamma, gammainc, kn, hyp2f1

import numdifftools as ndt
import emcee
from multiprocessing import Pool
from multiprocessing import cpu_count

ncpu = cpu_count()

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Peak position"
# ---------------------------------------------------------------------------


def gauss_sig(x_axis, gauss, peak):
    """
    This function estimates the dispersion of a Gaussian clump.

    Parameters
    ----------
    x_axis : array_like
        Array containing values from one particular direction.
    gauss : array_like
        Array containing the PDF at the repective values
        from one particular direction.
    peak : float
        Peak of the PDF, locally.

    Returns
    -------
    sigma : float
        Guess of the Gaussian dispersion.

    """

    # Gets the array's index where the peak is located
    i, j = np.argmin(np.abs(x_axis - peak)), np.argmin(np.abs(x_axis - peak))

    # Gets the histogram max value times e^(-1/2): (histogram value at 1 sigma)
    threshold = gauss[i] * np.exp(-1 / 2)

    # Search where the histogram is less than the threshold
    while gauss[i] >= threshold:
        i -= 1
    index_left = i
    while gauss[j] >= threshold:
        j += 1
    index_right = j

    # Assigns sigma as the minimum between the distance took to depass the
    # threshold in the left and in the right of the peak
    sigma = min(x_axis[index_right] - peak, peak - x_axis[index_left])

    return sigma


def find_center(x, y, method="mle", ra0=None, dec0=None, hybrid=True):
    """
    Fit a center (peak) of the [x,y] data.

    Parameters
    ----------
    x : array_like
        Data in x-direction
    y : array_like
        Data in y-direction
    method : string, optional
        Method to find the peak of RA, Dec. Available options are:
            - 'iterative'.
            - 'mle'.
            - 'mle_robust'.
        Default is 'mle'.
    ra0 : float, optional
        RA center of the original data set downloaded.
    dec0 : float, optional
        Dec center of the original data set downloaded.
    hydrid :  boolean, optional
        "True", if the user whises to consider field stars in the fit.
        The default is True.

    Raises
    ------
    ValueError
        Method argument is not one of the following:
            - 'kde'
            - 'iterative'

    Returns
    -------
    center : 2D-array of floats
        Position of the peak: [x_coordinate,y_coordinate]
    unc : float
        Uncertainty in the position.

    """

    if method not in ["iterative", "mle", "mle_robust"]:
        raise ValueError("Does not recognize method argument.")

    # Takes off NaN values
    x_nan = np.logical_not(np.isnan(x))
    y_nan = np.logical_not(np.isnan(y))
    idx_nan = x_nan * y_nan
    x = x[idx_nan]
    y = y[idx_nan]

    if method == "iterative":
        center, unc = center_iterative(x, y)
    elif method == "mle":
        center, unc = center_mle(x, y, hybrid=hybrid)
    elif method == "mle_robust":
        center, unc = center_mle_rob(x, y, ra0=ra0, dec0=dec0, hybrid=hybrid)

    return center, unc


def center_iterative(x, y):
    """
    Fit a center (peak) of the [x,y] data through an iterative approach.

    Parameters
    ----------
    x : array_like
        Data in x-direction
    y : array_like
        Data in y-direction

    Returns
    -------
    center : 2D-array of floats
        Position of the peak: [x_coordinate,y_coordinate]
    unc : float
        Uncertainty in the position.

    """

    bins_x = good_bin(x)
    bins_y = good_bin(y)

    # Gets the histogram in RA
    x_hist, x_axis = np.histogram(x, bins=bins_x, range=(np.amin(x), np.amax(x)))
    x_axis = 0.5 * (x_axis[1:] + x_axis[:-1])

    # Gets the histogram in Dec
    y_hist, y_axis = np.histogram(y, bins=bins_y, range=(np.amin(y), np.amax(y)))
    y_axis = 0.5 * (y_axis[1:] + y_axis[:-1])

    # Gets the histogram of the 2d (RA,Dec) data
    hist, xedges, yedges = np.histogram2d(x, y, bins=[bins_x, bins_y])
    hist = hist.T

    # Estimates the RA and Dec means from the galactic object by taking the
    # main local maxima of the 2d histogram.
    peaks = peak_local_max(hist, num_peaks=1)

    y_peak, x_peak = peaks.T[0], peaks.T[1]
    cmx, cmy = xedges[x_peak], yedges[y_peak]

    sigma_x = np.asarray([gauss_sig(x_axis, x_hist, cmx)])
    sigma_y = np.asarray([gauss_sig(y_axis, y_hist, cmy)])

    sigma = np.nanmean(np.asarray([sigma_x, sigma_y]))

    center = np.asarray([cmx, cmy])
    unc = np.nan

    many_tracers = True
    shift = list()
    count = 0
    while many_tracers is True:

        idx = np.where(
            angle.sky_distance_deg(x, y, center[0], center[1]) < (0.9 ** count) * sigma
        )

        bins_x = good_bin(x[idx])
        bins_y = good_bin(y[idx])

        if len(idx[0]) < bins_x * bins_y and count > 0:

            many_tracers = False
            return center, unc

        hist, xedges, yedges = np.histogram2d(x[idx], y[idx], bins=[bins_x, bins_y])

        xedges = 0.5 * (xedges[1:] + xedges[:-1])
        yedges = 0.5 * (yedges[1:] + yedges[:-1])

        cmx = np.nansum(xedges * np.sum(hist, axis=1)) / np.nansum(np.sum(hist, axis=1))
        cmy = np.nansum(yedges * np.sum(hist, axis=0)) / np.nansum(np.sum(hist, axis=0))

        shift.append(angle.sky_distance_deg(cmx, cmy, center[0], center[1]))

        center = np.asarray([cmx, cmy])
        unc = np.median(shift)

        count += 1

    return center, unc


def center_mle(x, y, hybrid=True):
    """
    Fit a center (peak) of the [x,y] data through a simple mle approach.

    Parameters
    ----------
    x : array_like
        Data in x-direction
    y : array_like
        Data in y-direction
    hydrid :  boolean, optional
        "True", if the user whises to consider field stars in the fit.
        The default is True.

    Returns
    -------
    center : 2D-array of floats
        Position of the peak: [x_coordinate,y_coordinate]
    unc : float
        Uncertainty in the position.

    """

    cmx, cmy = quantile(x, 0.5), quantile(y, 0.5)
    hmr, norm = initial_guess_sd(x=x, y=y, x0=cmx, y0=cmy)

    hmr = np.log10(hmr)
    norm = np.log10(norm)
    if hybrid is False:
        norm = -50

    bounds = [
        (hmr - 2, hmr + 2),
        (norm - 2, norm + 2),
        (quantile(x, 0.16), quantile(x, 0.84)),
        (quantile(y, 0.16), quantile(y, 0.84)),
    ]

    mle_model = differential_evolution(
        lambda c: likelihood_plummer_freec(c, x, y), bounds
    )
    results = mle_model.x
    hfun = ndt.Hessian(lambda c: likelihood_plummer_freec(c, x, y), full_output=True)

    hessian_ndt, info = hfun(results)
    if hybrid is False:
        arg_null = np.argmin(np.abs(np.diag(hessian_ndt)))
        hessian_ndt = np.delete(hessian_ndt, arg_null, axis=1)
        hessian_ndt = np.delete(hessian_ndt, arg_null, axis=0)
        var = np.sqrt(np.diag(np.linalg.inv(hessian_ndt)))

        center = np.asarray([results[1], results[2]])
        unc = np.asarray([var[1], var[2]])
    else:
        var = np.sqrt(np.diag(np.linalg.inv(hessian_ndt)))

        center = np.asarray([results[2], results[3]])
        unc = np.asarray([var[2], var[3]])

    return center, unc


def center_mle_rob(x, y, ra0=None, dec0=None, hybrid=True):
    """
    Fit a center (peak) of the [x,y] data through an mle robust approach.
    It considers the circular section where the data is complete.

    Parameters
    ----------
    x : array_like
        Data in x-direction
    y : array_like
        Data in y-direction
    ra0 : float, optional
        RA center of the original data set downloaded.
    dec0 : float, optional
        Dec center of the original data set downloaded.
    hydrid :  boolean, optional
        "True", if the user whises to consider field stars in the fit.
        The default is True.

    Returns
    -------
    center : 2D-array of floats
        Position of the peak: [x_coordinate,y_coordinate]
    unc : float
        Uncertainty in the position.

    """
    if ra0 is None:
        ra0 = 0.5 * (max(x) + min(x))
    if dec0 is None:
        dec0 = 0.5 * (max(y) + min(y))
    rmax = np.nanmax(angle.sky_distance_deg(x, y, ra0, dec0))

    cmx, cmy = ra0, dec0
    hmr, norm = initial_guess_sd(x=x, y=y, x0=cmx, y0=cmy)

    hmr = np.log10(hmr)
    norm = np.log10(norm)
    if hybrid is False:
        norm = -50

    bounds = [
        (hmr - 2, hmr + 2),
        (norm - 2, norm + 2),
        (quantile(x, 0.16), quantile(x, 0.84)),
        (quantile(y, 0.16), quantile(y, 0.84)),
    ]

    mle_model = differential_evolution(
        lambda c: likelihood_plummer_center(c, x, y, ra0, dec0, rmax), bounds
    )
    results = mle_model.x
    hfun = ndt.Hessian(
        lambda c: likelihood_plummer_center(c, x, y, ra0, dec0, rmax), full_output=True
    )

    hessian_ndt, info = hfun(results)
    if hybrid is False:
        arg_null = np.argmin(np.abs(np.diag(hessian_ndt)))
        hessian_ndt = np.delete(hessian_ndt, arg_null, axis=1)
        hessian_ndt = np.delete(hessian_ndt, arg_null, axis=0)

        var = np.sqrt(np.diag(np.linalg.inv(hessian_ndt)))

        center = np.asarray([results[1], results[2]])
        unc = np.asarray([var[1], var[2]])
    else:
        var = np.sqrt(np.diag(np.linalg.inv(hessian_ndt)))

        center = np.asarray([results[2], results[3]])
        unc = np.asarray([var[2], var[3]])

    return center, unc


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Surface density"
# ---------------------------------------------------------------------------


def surface_density(x=None, y=None, x0=None, y0=None):
    """
    Binned surface density derivation.

    Parameters
    ----------
    x : array_like, optional
        Data in x-direction. The default is None.
    y : array_like, optional
        Data in x-direction. The default is None.
    x0 : float, optional
        Peak of data in x-direction. The default is None.
    y0 : TYPE, optional
        Peak of data in y-direction. The default is None.

    Raises
    ------
    ValueError
        No data is provided.

    Returns
    -------
    density : array_like
        Binned array containing, at positions:
            0 : Projected radius.
            1 : Surface density.
            2 : Poisson error on surface density.
            3 : Size inbetween projected radii.

    """

    if (x is None and y is None) or (x is None):
        raise ValueError("Please provide the data to be fitted.")

    if y is None:
        r = x
    else:

        if x0 is None or y0 is None:

            center, unc = find_center(x, y)
            if x0 is None:
                x0 = center[0]
            if y0 is None:
                y0 = center[1]

        r = angle.sky_distance_deg(x, y, x0, y0)

    q_16, q_50, q_84 = quantile(r, [0.16, 0.5, 0.84])
    q_m, q_p = q_50 - q_16, q_84 - q_50
    nbins = int((np.amax(r) - np.amin(r)) / (min(q_m, q_p) / 5))

    idx_valid = np.where(r > 0)

    counts, binlim = np.histogram(
        r,
        range=(np.nanmin(r[idx_valid]), np.nanmax(r[idx_valid])),
        bins=np.logspace(
            np.nanmin(np.log10(r[idx_valid])),
            np.nanmax(np.log10(r[idx_valid])),
            nbins + 1,
        ),
    )

    bincent = 0.5 * (binlim[1:] + binlim[:-1])
    barsize = binlim[1:] - binlim[:-1]
    surface = np.pi * (binlim[1:] ** 2 - binlim[:-1] ** 2)
    s_dens = counts / surface
    errs = np.zeros(len(counts))
    for i in range(len(counts)):
        if counts[i] == 0:
            errs[i] = np.nan
        else:
            errs[i] = s_dens[i] / np.sqrt(counts[i])

    s_dens[np.where(counts == 0)] = np.nan

    density = np.zeros((4, nbins))
    density[0] = bincent
    density[1] = s_dens
    density[2] = errs
    density[3] = barsize

    return density


def initial_guess_sd(x=None, y=None, x0=None, y0=None):
    """
    Initial guess of projected half mass radius.

    Parameters
    ----------
    x : array_like, optional
        Data in x-direction. The default is None.
    y : array_like, optional
        Data in x-direction. The default is None.
    x0 : float, optional
        Peak of data in x-direction. The default is None.
    y0 : float, optional
        Peak of data in y-direction. The default is None.

    Raises
    ------
    ValueError
        No data is provided.

    Returns
    -------
    half_radius : float
        Guess on projected half mass radius.

    """
    if (x is None and y is None) or (x is None):
        raise ValueError("Please provide the data to be fitted.")

    if y is None:
        density = surface_density(x=x)
        size = len(x[0])
    else:

        if x0 is None or y0 is None:

            center, unc = find_center(x, y)
            if x0 is None:
                x0 = center[0]
            if y0 is None:
                y0 = center[1]

        density = surface_density(x=x, y=y, x0=x0, y0=y0)
        size = len(x)

    mw_dens = np.nanmin(density[1])
    density[1] = density[1] - mw_dens
    nilop = (
        mw_dens * np.pi * (density[0][len(density[0]) - 1] ** 2 - density[0][0] ** 2)
    )
    n_local = np.zeros(len(density[0]))
    for i in range(len(density[0])):
        n_local[i] = (
            np.nansum(n_local)
            + 2 * np.pi * density[1][i] * density[0][i] * density[3][i]
        )

    idx = np.nanargmin(np.abs(n_local - n_local[len(n_local) - 1] * 0.5))

    half_radius = density[0][idx]
    norm = size / (size - nilop) - 1
    if norm <= 0:
        norm = 1

    return half_radius, norm


def prob(r, params, model="plummer"):
    """
    Computes the membership probability of the stars based
    on the surface density fits alone.

    Parameters
    ----------
    r : array_like, optional
        Projected radii (in same units as the scale radius).
    params : array_like
        Parameters to be fitted: Output of the maximum_likelihood method.
    model : string, optional
        Surface density model to be considered. Available options are:
             - 'sersic'
             - 'kazantzidis'
             - 'plummer'
        The default is 'plummer'.

    Raises
    ------
    ValueError
        Surface density model is not one of the following:
            - 'sersic'
            - 'kazantzidis'
            - 'plummer'

    Returns
    -------
    probability : array_like
        Probability of a each star to belong to the respective
        a galactic object (considering only proper motions).

    """

    if model not in ["sersic", "plummer", "kazantzidis"]:
        raise ValueError("Does not recognize surface density model.")

    if model == "plummer" or model == "kazantzidis":
        n = 0
        a = 10 ** params[0]
        if params[1] < -10:
            norm = 0
        else:
            norm = 10 ** params[1]
    elif model == "sersic":
        n = params[0]
        a = 10 ** params[1]
        if params[2] < -10:
            norm = 0
        else:
            norm = 10 ** params[2]

    nsys = len(r) / (1 + norm)
    nilop = len(r) - nsys

    Xmax = np.amax(r) / a
    Xmin = np.amin(r) / a
    X = r / a

    if model == "plummer":
        sd = sd_plummer(X) * nsys / (np.pi * a ** 2)
    elif model == "kazantzidis":
        sd = sd_kazantzidis(X) * nsys / (np.pi * a ** 2)
    elif model == "sersic":
        sd = sd_sersic(n, X) * nsys / (np.pi * a ** 2)

    sd_fs = nilop / (np.pi * (Xmax ** 2 - Xmin ** 2))

    probability = sd / (sd + sd_fs)

    return probability


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Sersic profile functions"
# ---------------------------------------------------------------------------

###############################################################################
#
# Functions concerning the Sersic profile (Sersic 1963, 1968).
#
###############################################################################


def b(n):
    """
    Gets the Sersic's b_n term, using the approximation
    from Ciotti & Bertin (1999).

    Parameters
    ----------
    n : array_like, float
        Sersic index

    Returns
    -------
    b : array_like, float
        Sersic's b_n term.

    """

    b = (
        2 * n
        - 1 / 3
        + 4 / (405 * n)
        + 46 / (25515 * n ** 2)
        + 131 / (1148175 * n ** 3)
        - 2194697 / (30690717750 * n ** 4)
    )
    return b


def sd_sersic(n, X):
    """
    Sersic surface density, normalized according to the convention:

    SD(X = R/R_e) = SD_real(R) * pi * R_e^2 / N_infinty

    Parameters
    ----------
    n : array_like, float
        Sersic index.
    X : array_like (same shape as n), float
        Projected radius X = R/R_e.

    Returns
    -------
    SD : array_like (same shape as n), float
        Normalized surface density profile.

    """

    bn = b(n)
    sd = np.exp(-bn * X ** (1 / n)) * (bn ** (2 * n) / (2 * n * gamma(2 * n)))
    return sd


def n_sersic(n, X):
    """
    Sersic projected number, normalized according to the convention:

    N(X = R/R_e) = N(R) / N_infinty

    Parameters
    ----------
    n : array_like, float
        Sersic index.
    X : array_like (same shape as n), float
        Projected radius X = R/R_e.

    Returns
    -------
    N : array_like (same shape as n), float
        Sersic projected number.

    """

    bn = b(n)
    N = gammainc(2 * n, bn * X ** (1 / n))
    return N


def likelihood_sersic(params, Ri):
    """
    Likelihood function of the Sersic profile plus a constant contribution
    from fore/background tracers.

    Parameters
    ----------
    Parameters to be fitted: Sersic index, Sersic characteristic radius R_e and
                             log-ratio of galactic objects and Milky
                             Way stars.
    Ri : array_like
        Array containing the ensemble of projected radii.

    Returns
    -------
    L : float
       Likelihood.

    """

    n = params[0]
    Re = 10 ** params[1]
    if params[2] < -10:
        norm = 0
    else:
        norm = 10 ** params[2]

    Xmax = np.amax(Ri) / Re
    Xmin = np.amin(Ri) / Re
    X = Ri / Re

    N_sys_tot = n_sersic(n, Xmax) - n_sersic(n, Xmin)

    SD = sd_sersic(n, X) + norm * N_sys_tot / (Xmax ** 2 - Xmin ** 2)

    Ntot = N_sys_tot * (1 + norm)

    fi = 2 * (X / Re) * SD / Ntot

    idx_valid = np.logical_not(np.isnan(np.log(fi)))

    L = -np.sum(np.log(fi[idx_valid]))

    return L


def lnprior_s(params, guess, bounds):
    """
    Prior assumptions on the parameters.

    Parameters
    ----------
    Parameters to be fitted: Sersic index, Sersic characteristic radius R_e and
                             log-ratio of galactic objects and Milky
                             Way stars.
    guess : array_like
        Array containing the initial guess of the parameters.
    bounds : array_like
        Array containing the interval of variation of the parameters.


    Returns
    -------
    log-prior probability : float
        0, if the parameters are within the prior range,
        - Infinity otherwise.

    """

    if (
        (guess[0] - bounds[0] <= params[0] <= guess[0] + bounds[0])
        and (guess[1] - bounds[1] <= params[1] <= guess[1] + bounds[1])
        and (guess[2] - bounds[2] <= params[2] <= guess[2] + bounds[2])
    ):
        return 0.0
    return -np.inf


def lnprob_s(params, Ri, guess, bounds):
    """
    log-probability of fit parameters.

    Parameters
    ----------
    Parameters to be fitted: Sersic index, Sersic characteristic radius R_e and
                             log-ratio of galactic objects and Milky
                             Way stars.
    Ri : array_like
        Array containing the ensemble of projected radii.
    guess : array_like
        Array containing the initial guess of the parameters.
    bounds : array_like
        Array containing the interval of variation of the parameters.


    Returns
    -------
    log-prior probability : float
        log-probability of fit parameters.

    """

    lp = lnprior_s(params, guess, bounds)
    if not np.isfinite(lp):
        return -np.inf
    return lp - likelihood_sersic(params, Ri)


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"gPlummer profile functions"
# ---------------------------------------------------------------------------

###############################################################################
#
# Functions concerning the general Plummer profile
# (see http://www2.iap.fr/users/gam/mkmocks.pdf).
#
###############################################################################


def sd_gplummer(gam, X):
    """
    gPlummer surface density, normalized according to the convention:

    SD(X = R/a) = SD_real(R) * pi * a^2 / N_infinty

    Parameters
    ----------
    gam : array_like, float
        Inner slope.
    X : array_like (same shape as gam), float
        Projected radius X = R/a.

    Returns
    -------
    SD : array_like (same shape as gam), float
        Normalized surface density profile.

    """

    term1 = (
        -(1 + X * X)
        * (-3 + X * X + gam)
        * hyp2f1(1, (5 - gam) / 2, -1 / 2, -1 / (X * X))
    )
    term2 = (-17 + X ** 4 - X * X * (gam - 8) - gam * (gam - 9)) * hyp2f1(
        1, (5 - gam) / 2, 1 / 2, -1 / (X * X)
    )
    term3 = (gam - 3 - X * X) * (term1 + term2)

    num = -(gam - 3) * (X * X * (gam - 4) * (gam - 2) + term3)
    den = 2 * X ** 4 * (1 + X * X) * (gam - 4) * (gam - 2)

    sd = num / den

    return sd


def n_gplummer(gam, X):
    """
    gPlummer projected number, normalized according to the convention:

    N(X = R/a) = N(R) / N_infinty

    Parameters
    ----------
    gam : array_like, float
        Inner slope.
    X : array_like (same shape as gam), float
        Projected radius X = R/a.

    Returns
    -------
    N : array_like (same shape as gam), float
        gPlummer projected number.

    """

    term1 = 1 / (X * X * (gam - 4) * (gam - 2))
    term2 = (
        (1 + X * X) * (X * X + gam - 3) * hyp2f1(1, (5 - gam) / 2, -1 / 2, -1 / (X * X))
    )
    term3 = (17 - X ** 4 + X * X * (gam - 8) + gam * (gam - 9)) * hyp2f1(
        1, (5 - gam) / 2, 1 / 2, -1 / (X * X)
    )

    N = 1 + term1 * (term2 + term3) * (3 - gam)

    return N


def likelihood_gplummer(params, Ri):
    """
    Likelihood function of the gPlummer profile plus a constant contribution
    from fore/background tracers.

    Parameters
    ----------
    Parameters to be fitted: gPlummer inner slope, gPlummer characteristic radius a and
                             log-ratio of galactic objects and Milky
                             Way stars.
    Ri : array_like
        Array containing the ensemble of projected radii.

    Returns
    -------
    L : float
       Likelihood.

    """

    gam = params[0]
    a = 10 ** params[1]
    if params[2] < -10:
        norm = 0
    else:
        norm = 10 ** params[2]

    Xmax = np.amax(Ri) / a
    Xmin = np.amin(Ri) / a
    X = Ri / a

    N_sys_tot = n_gplummer(gam, Xmax) - n_gplummer(gam, Xmin)

    SD = sd_gplummer(gam, X) + norm * N_sys_tot / (Xmax ** 2 - Xmin ** 2)

    Ntot = N_sys_tot * (1 + norm)

    fi = 2 * (X / a) * SD / Ntot

    idx_valid = np.logical_not(np.isnan(np.log(fi)))

    L = -np.sum(np.log(fi[idx_valid]))

    return L


def lnprior_gp(params, guess, bounds):
    """
    Prior assumptions on the parameters.

    Parameters
    ----------
    Parameters to be fitted: gPlummer inner slope, gPlummer characteristic radius a and
                             log-ratio of galactic objects and Milky
                             Way stars.
    guess : array_like
        Array containing the initial guess of the parameters.
    bounds : array_like
        Array containing the interval of variation of the parameters.


    Returns
    -------
    log-prior probability : float
        0, if the parameters are within the prior range,
        - Infinity otherwise.

    """

    if (
        (guess[0] - bounds[0] <= params[0] <= guess[0] + bounds[0])
        and (guess[1] - bounds[1] <= params[1] <= guess[1] + bounds[1])
        and (guess[2] - bounds[2] <= params[2] <= guess[2] + bounds[2])
    ):
        return 0.0
    return -np.inf


def lnprob_gp(params, Ri, guess, bounds):
    """
    log-probability of fit parameters.

    Parameters
    ----------
    Parameters to be fitted: gPlummer inner slope, gPlummer characteristic radius a and
                             log-ratio of galactic objects and Milky
                             Way stars.
    Ri : array_like
        Array containing the ensemble of projected radii.
    guess : array_like
        Array containing the initial guess of the parameters.
    bounds : array_like
        Array containing the interval of variation of the parameters.


    Returns
    -------
    log-prior probability : float
        log-probability of fit parameters.

    """

    lp = lnprior_s(params, guess, bounds)
    if not np.isfinite(lp):
        return -np.inf
    return lp - likelihood_sersic(params, Ri)


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Kazantzidis profile functions"
# ---------------------------------------------------------------------------

###############################################################################
#
# Functions concerning the Kazantzidis profile (Kazantzidis et al. 2004).
#
###############################################################################


def sd_kazantzidis(X):
    """
    Kazantzidis surface density, normalized according to the convention:

    SD(X = R/a) = SD_real(R) * pi * a^2 / N_infinty

    Parameters
    ----------
    X : array_like (same shape as n), float
        Projected radius X = R/a

    Returns
    -------
    SD : array_like (same shape as X), float
        Normalized surface density profile.

    """

    sd = kn(0, X) / 2
    return sd


def n_kazantizidis(X):
    """
    Kazantzidis projected number, normalized according to the convention:

    N(X = R/a) = N(R) / N_infinty

    Parameters
    ----------
    X : array_like (same shape as n), float
        Projected radius X = R/a.

    Returns
    -------
    N : array_like (same shape as n), float
        Kazantzidis projected number.

    """

    N = 1 - X * kn(1, X)
    return N


def likelihood_kazantzidis(params, Ri):
    """
    Likelihood function of the Sersic profile plus a constant contribution
    from fore/background tracers.

    Parameters
    ----------
    Parameters to be fitted: Kazantzidis characteristic radius a and
                             log-ratio of galactic objects and Milky
                             Way stars.
    Ri : array_like
        Array containing the ensemble of projected radii.

    Returns
    -------
    L : float
       Likelihood.

    """

    a = 10 ** params[0]
    if params[1] < -10:
        norm = 0
    else:
        norm = 10 ** params[1]

    Xmax = np.amax(Ri) / a
    Xmin = np.amin(Ri) / a
    X = Ri / a

    N_sys_tot = n_kazantizidis(Xmax) - n_kazantizidis(Xmin)

    SD = sd_kazantzidis(X) + norm * N_sys_tot / (Xmax ** 2 - Xmin ** 2)

    Ntot = N_sys_tot * (1 + norm)

    fi = 2 * (X / a) * SD / Ntot

    idx_valid = np.logical_not(np.isnan(np.log(fi)))

    L = -np.sum(np.log(fi[idx_valid]))

    return L


def lnprior_k(params, guess, bounds):
    """
    Prior assumptions on the parameters.

    Parameters
    ----------
    Parameters to be fitted: Kazantzidis characteristic radius a and
                             log-ratio of galactic objects and Milky
                             Way stars.
    guess : array_like
        Array containing the initial guess of the parameters.
    bounds : array_like
        Array containing the interval of variation of the parameters.

    Returns
    -------
    log-prior probability : float
        0, if the parameters are within the prior range,
        - Infinity otherwise.

    """

    if (guess[0] - bounds[0] <= params[0] <= guess[0] + bounds[0]) and (
        guess[1] - bounds[1] <= params[1] <= guess[1] + bounds[1]
    ):
        return 0.0
    return -np.inf


def lnprob_k(params, Ri, guess, bounds):
    """
    log-probability of fit parameters.

    Parameters
    ----------
        Parameters to be fitted: Kazantzidis characteristic radius a and
                                 log-ratio of galactic objects and Milky
                                 Way stars.
    Ri : array_like
        Array containing the ensemble of projected radii.
    guess : array_like
        Array containing the initial guess of the parameters.
    bounds : array_like
        Array containing the interval of variation of the parameters.

    Returns
    -------
    log-prior probability : float
        log-probability of fit parameters.

    """

    lp = lnprior_k(params, guess, bounds)
    if not np.isfinite(lp):
        return -np.inf
    return lp - likelihood_kazantzidis(params, Ri)


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Plummer profile functions"
# ---------------------------------------------------------------------------

###############################################################################
#
# Functions concerning the Plummer profile (Plummer 1911).
#
###############################################################################


def sd_plummer(X):
    """
    Plummer surface density, normalized according to the convention:

    SD(X = R/a) = SD_real(R) * pi * a^2 / N_infinty

    Parameters
    ----------
    X : array_like (same shape as n), float
        Projected radius X = R/a

    Returns
    -------
    SD : array_like (same shape as X), float
        Normalized surface density profile.

    """

    sd = 1 / ((1 + X * X) * (1 + X * X))
    return sd


def n_plummer(X):
    """
    Plummer projected number, normalized according to the convention:

    N(X = R/a) = N(R) / N_infinty

    Parameters
    ----------
    X : array_like (same shape as n), float
        Projected radius X = R/a.

    Returns
    -------
    N : array_like (same shape as a), float
        Plummer projected number.

    """

    N = X * X / (1 + X * X)
    return N


def n_plummer_angle(R, Rmax, d, a):
    """
    Returns the angle analyzed in the surface density fit, for each value of R.

    Parameters
    ----------
    R : array_like
        Distance from center in degrees.
    Rmax : float
        Maximum radius of original data set.
    d : float
        Distance from original center.
    a : float
        Plummer scale radius.

    Returns
    -------
    N : array_like (same shape as n), float
        Plummer projected number.

    """

    if R <= Rmax - d:
        return n_plummer(R / a)
    else:
        if R >= d + Rmax:
            return 1
        arg1 = (R * R + d * d - Rmax * Rmax) / (2 * R * d)
        term1 = -a * a * np.arccos(arg1) / (np.pi * (a * a + R * R))

        arg2 = np.sqrt(
            -(d ** 4) - (R * R - Rmax * Rmax) ** 2 + 2 * d * d * (R * R + Rmax * Rmax)
        )

        arg3 = -np.sqrt(
            a ** 4 + (d * d - Rmax * Rmax) ** 2 + 2 * a * a * (d * d + Rmax * Rmax)
        )

        arg4 = (d * d - Rmax * Rmax) ** 2 - R * R * (d * d + Rmax * Rmax)

        arg5 = np.arctan(arg4 / (arg2 * (d * d - Rmax * Rmax)))

        term2 = arg3 * arg5

        arg6 = a * a + d * d - Rmax * Rmax

        arg7 = (
            (d * d - Rmax * Rmax) ** 2
            + a * a * (d * d + Rmax * Rmax)
            - R * R * (a * a + d * d + Rmax * Rmax)
        )

        arg8 = arg6 * np.arctan(arg7 / (-arg3 * arg2))

        term3 = (term2 + arg8) * arg2

        arg9 = -2 * np.pi * arg3 * arg2

        term4 = term3 / arg9

        n = 1 + (term1 + term4)

        return n


def angle_section(R, Rmax, d):
    """
    Returns the angle of the circular section analyzed in the
    surface density fit, for each value of R.

    Parameters
    ----------
    R : array_like
        Distance from center in degrees.
    Rmax : float
        Maximum radius of original data set.
    d : float
        Distance from original center.

    Returns
    -------
    phi : array_like
        angle analyzed in the surface density fit, for each value of R.

    """

    phi = np.zeros(len(R))

    idx_phi = np.where(R > Rmax - d)
    idx_2pi = np.where(R <= Rmax - d)

    argphi = (R[idx_phi] * R[idx_phi] + d * d - Rmax * Rmax) / (2 * R[idx_phi] * d)

    phi[idx_phi] = 2 * np.arccos(argphi)
    phi[idx_2pi] = 2 * np.pi * np.ones(len(R[idx_2pi]))

    return phi


def likelihood_plummer(params, Ri):
    """
    Likelihood function of the Plummer profile plus a constant contribution
    from fore/background tracers.

    Parameters
    ----------
    params : array_like
        Parameters to be fitted: Plummer characteristic radius a and
                                 log-ratio of galactic objects and Milky
                                 Way stars.
    Ri : array_like
        Array containing the ensemble of projected radii.

    Returns
    -------
    L : float
       Likelihood.

    """

    a = 10 ** params[0]
    if params[1] < -10:
        norm = 0
    else:
        norm = 10 ** params[1]

    Xmax = np.amax(Ri) / a
    Xmin = np.amin(Ri) / a
    X = Ri / a

    N_sys_tot = n_plummer(Xmax) - n_plummer(Xmin)

    SD = sd_plummer(X) + norm * N_sys_tot / (Xmax ** 2 - Xmin ** 2)

    Ntot = N_sys_tot * (1 + norm)

    fi = 2 * (X / a) * SD / Ntot

    idx_valid = np.logical_not(np.isnan(np.log(fi)))

    L = -np.sum(np.log(fi[idx_valid]))

    return L


def likelihood_plummer_freec(params, x, y):
    """
    Likelihood function of the Plummer profile plus a constant contribution
    from fore/background tracers. Differently from the previous method,
    it also fits the center of the distribution.

    Parameters
    ----------
    params : array_like
        Parameters to be fitted: Plummer characteristic radius a and
                                 log-ratio of galactic objects and Milky
                                 Way stars.
    x : array_like
        Array containing the ensemble of ra data.
    y : array_like
        Array containing the ensemble of dec data.

    Returns
    -------
    L : float
       Likelihood.

    """

    a = 10 ** params[0]
    if params[1] < -10:
        norm = 0
    else:
        norm = 10 ** params[1]

    cmx = params[2]
    cmy = params[3]

    Ri = angle.sky_distance_deg(cmx, cmy, x, y)

    Xmax = np.amax(Ri) / a
    Xmin = np.amin(Ri) / a
    X = Ri / a

    N_sys_tot = n_plummer(Xmax) - n_plummer(Xmin)

    SD = sd_plummer(X) + norm * N_sys_tot / (Xmax ** 2 - Xmin ** 2)

    Ntot = N_sys_tot * (1 + norm)

    fi = 2 * (X / a) * SD / Ntot

    idx_valid = np.logical_not(np.isnan(np.log(fi)))

    L = -np.sum(np.log(fi[idx_valid]))

    return L


def likelihood_plummer_center(params, x, y, ra0, dec0, rmax):
    """
    Likelihood function of the Plummer profile plus a constant contribution
    from fore/background tracers. Differently from the previous method,
    it also fits the center of the distribution, but considering the circular
    section where the data is complete.

    Parameters
    ----------
    params : array_like
        Parameters to be fitted: Plummer characteristic radius a and
                                 log-ratio of galactic objects and Milky
                                 Way stars.
    x : array_like
        Array containing the ensemble of ra data.
    y : array_like
        Array containing the ensemble of dec data.
    ra0 : float
        RA center from SIMBAD
    dec0 : float
        Dec center from SIMBAD
    rmax : float
        Original maximum projected radius in the data.

    Returns
    -------
    L : float
       Likelihood.

    """

    a = 10 ** params[0]
    if params[1] < -10:
        norm = 0
    else:
        norm = 10 ** params[1]

    cmx = params[2]
    cmy = params[3]

    Ri = angle.sky_distance_deg(cmx, cmy, x, y)

    d = angle.sky_distance_deg(cmx, cmy, ra0, dec0)

    phi = angle_section(Ri, rmax, d)

    Xmax = np.amax(Ri) / a
    Xmin = np.amin(Ri) / a
    X = Ri / a

    N_sys_tot = n_plummer_angle(Xmax * a, rmax, d, a) - n_plummer_angle(
        Xmin * a, rmax, d, a
    )

    SD = sd_plummer(X) + norm * N_sys_tot / (rmax ** 2)

    Ntot = N_sys_tot * (1 + norm)

    fi = (phi / np.pi) * (X / a) * SD / Ntot

    idx_valid = np.logical_not(np.isnan(np.log(fi)))

    L = -np.sum(np.log(fi[idx_valid]))

    return L


def lnprior_p(params, guess, bounds):
    """
    Prior assumptions on the parameters.

    Parameters
    ----------
    params : array_like
        Parameters to be fitted: Plummer characteristic radius a and
                                 log-ratio of galactic objects and Milky
                                 Way stars.
    guess : array_like
        Array containing the initial guess of the parameters.
    bounds : array_like
        Array containing the interval of variation of the parameters.

    Returns
    -------
    log-prior probability : float
        0, if the parameters are within the prior range,
        - Infinity otherwise.

    """

    if (guess[0] - bounds[0] <= params[0] <= guess[0] + bounds[0]) and (
        guess[1] - bounds[1] <= params[1] <= guess[1] + bounds[1]
    ):
        return 0.0
    return -np.inf


def lnprob_p(params, Ri, guess, bounds):
    """
    log-probability of fit parameters.

    Parameters
    ----------
    params : array_like
    Parameters to be fitted: Plummer characteristic radius a and
                             log-ratio of galactic objects and Milky
                             Way stars.
    Ri : array_like
        Array containing the ensemble of projected radii.
    guess : array_like
        Array containing the initial guess of the parameters.
    bounds : array_like
        Array containing the interval of variation of the parameters.


    Returns
    -------
    log-prior probability : float
        log-probability of fit parameters.

    """

    lp = lnprior_p(params, guess, bounds)
    if not np.isfinite(lp):
        return -np.inf
    return lp - likelihood_plummer(params, Ri)


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"General functions"
# ---------------------------------------------------------------------------


def quantile(x, q):
    """
    Compute sample quantiles.

    Parameters
    ----------
    x : array_like
        Array containing set of values.
    q : array_like, float
        Quantile values to be derived (must be between 0 and 1).

    Raises
    ------
    ValueError
        Quantiles must be between 0 and 1.

    Returns
    -------
    percentile: array_like (same shape as q), float
        Percentile value(s).

    """

    x = np.atleast_1d(x)
    q = np.atleast_1d(q)

    if np.any(q < 0.0) or np.any(q > 1.0):
        raise ValueError("Quantiles must be between 0 and 1")

    return np.percentile(x, 100.0 * q)


def good_bin(x):
    """
    Computes an adequate number of bins.

    Parameters
    ----------
    x : array_like
        Array to be binned.

    Returns
    -------
    bins : float
        Number of bins.

    """

    q_16, q_50, q_84 = quantile(x, [0.16, 0.5, 0.84])
    q_m, q_p = q_50 - q_16, q_84 - q_50
    bins = int((np.amax(x) - np.amin(x)) / (min(q_m, q_p) / 4))

    return bins


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Fitting procedure"
# ---------------------------------------------------------------------------


def maximum_likelihood(x=None, y=None, model="plummer", x0=None, y0=None, hybrid=True):
    """
    Calls a maximum likelihood fit of the surface density paramters of
    the joint distribution of galactic object plus Milky Way stars.

    Parameters
    ----------
    x : array_like, optional
        Data in x-direction. The default is None.
    y : array_like, optional
        Data in x-direction. The default is None.
    model : string, optional
        Surface density model to be considered. Available options are:
             - 'sersic'
             - 'kazantzidis'
             - 'plummer'
             - 'gplummer'
        The default is 'plummer'.
    x0 : float, optional
        Peak of data in x-direction. The default is None.
    y0 : TYPE, optional
        Peak of data in y-direction. The default is None.
    hydrid :  boolean, optional
        "True", if the user whises to consider field stars in the fit.
        The default is True.

    Raises
    ------
    ValueError
        Surface density model is not one of the following:
            - 'sersic'
            - 'kazantzidis'
            - 'plummer'
            - 'gplummer'
        No data is provided.

    Returns
    -------
    results : array
        Best fit parameters of the surface density model.
    var : array
        Uncertainty of the fits.

    """

    if model not in ["sersic", "plummer", "kazantzidis", "gplummer"]:
        raise ValueError("Does not recognize surface density model.")

    if (x is None and y is None) or (x is None):
        raise ValueError("Please provide the data to be fitted.")

    if y is None:
        ri = x
        hmr, norm = initial_guess_sd(x=x)
    else:

        if x0 is None or y0 is None:

            center, unc = find_center(x, y)
            if x0 is None:
                x0 = center[0]
            if y0 is None:
                y0 = center[1]
        ri = np.asarray([angle.sky_distance_deg(x, y, x0, y0)])
        hmr, norm = initial_guess_sd(x=x, y=y, x0=x0, y0=y0)

    hmr = np.log10(hmr)
    norm = np.log10(norm)
    if hybrid is False:
        norm = -50

    if model == "sersic":
        bounds = [(0.5, 10), (hmr - 2, hmr + 2), (norm - 2, norm + 2)]
        mle_model = differential_evolution(lambda c: likelihood_sersic(c, ri), bounds)
        results = mle_model.x
        hfun = ndt.Hessian(lambda c: likelihood_sersic(c, ri), full_output=True)

    elif model == "kazantzidis":
        bounds = [(hmr - 2, hmr + 2), (norm - 2, norm + 2)]
        mle_model = differential_evolution(
            lambda c: likelihood_kazantzidis(c, ri), bounds
        )
        results = mle_model.x
        hfun = ndt.Hessian(lambda c: likelihood_kazantzidis(c, ri), full_output=True)

    elif model == "plummer":
        bounds = [(hmr - 2, hmr + 2), (norm - 2, norm + 2)]
        mle_model = differential_evolution(lambda c: likelihood_plummer(c, ri), bounds)
        results = mle_model.x
        hfun = ndt.Hessian(lambda c: likelihood_plummer(c, ri), full_output=True)

    elif model == "gplummer":
        bounds = [(0, 2), (hmr - 2, hmr + 2), (norm - 2, norm + 2)]
        mle_model = differential_evolution(lambda c: likelihood_gplummer(c, ri), bounds)
        results = mle_model.x
        hfun = ndt.Hessian(lambda c: likelihood_gplummer(c, ri), full_output=True)

    hessian_ndt, info = hfun(results)
    if hybrid is False:
        arg_null = np.argmin(np.abs(np.diag(hessian_ndt)))
        hessian_ndt = np.delete(hessian_ndt, arg_null, axis=1)
        hessian_ndt = np.delete(hessian_ndt, arg_null, axis=0)

    var = np.sqrt(np.diag(np.linalg.inv(hessian_ndt)))

    return results, var


def mcmc(
    x=None,
    y=None,
    model="plummer",
    nwalkers=None,
    steps=1000,
    ini=None,
    bounds=None,
    use_pool=False,
    x0=None,
    y0=None,
    hybrid=True,
):
    """
    MCMC routine based on the emcee package (Foreman-Mackey et al, 2013).

    The user is strongly encouraged to provide initial guesses,
    which can be derived with the "maximum_likelihood" method,
    previously checked to provide reasonable fits.

    Parameters
    ----------
    x : array_like
        Data in x-direction.
    y : array_like
        Data in y-direction.
    model : string, optional
        Surface density model to be considered. Available options are:
             - 'sersic'
             - 'kazantzidis'
             - 'plummer'
             - 'gplummer'
        The default is 'plummer'.
    nwalkers : int, optional
        Number of Markov chains. The default is None.
    steps : int, optional
        Number of steps for each chain. The default is 1000.
    ini : array_like, optional
        Array containing the initial guess of the parameters.
        The order of parameters should be the same returned by the method
        "maximum_likelihood".
        The default is None.
    bounds : array_like, optional
        Array containing the allowed variation of the parameters, with
        respect to the initial guesses.
        The order of parameters should be the same returned by the method
        "maximum_likelihood".
        The default is None.
    use_pool : boolean, optional
        "True", if the user whises to use full CPU power of the machine.
        The default is False.
    x0 : float, optional
        Peak of data in x-direction. The default is None.
    y0 : float, optional
        Peak of data in y-direction. The default is None.
    hydrid :  boolean, optional
        "True", if the user whises to consider field stars in the fit.
        The default is True.

    Raises
    ------
    ValueError
        Surface density model is not one of the following:
            - 'sersic'
            - 'kazantzidis'
            - 'plummer'
            - 'gplummer'
        No data is provided.

    Returns
    -------
    chain : array_like
        Set of chains from the MCMC.

    """

    if model not in ["sersic", "plummer", "kazantzidis", "gplummer"]:
        raise ValueError("Does not recognize surface density model.")

    if (x is None and y is None) or (x is None):
        raise ValueError("Please provide the data to be fitted.")

    if y is None:
        ri = x
        if ini is None:
            ini, var = maximum_likelihood(x=x, model=model)
    else:

        if x0 is None or y0 is None:

            center, unc = find_center(x, y)
            if x0 is None:
                x0 = center[0]
            if y0 is None:
                y0 = center[1]
        ri = np.asarray([angle.sky_distance_deg(x, y, x0, y0)])

        if ini is None:
            ini, var = maximum_likelihood(x=x, y=y, x0=x0, y0=y0, model=model)
            if model == "sersic":
                ini = np.asarray([2, ini[0], ini[1]])
            if model == "gplummer":
                ini = np.asarray([1, ini[0], ini[1]])

    ndim = len(ini)  # number of dimensions.
    if nwalkers is None or nwalkers < 2 * ndim:
        nwalkers = int(2 * ndim + 1)

    if bounds is None:
        bounds = 3 * var

    if hybrid is False:
        if model == "sersic" or model == "gplummer":
            ini[2] = -50
        else:
            ini[1] = -50

    pos = [ini + 1e-3 * bounds * np.random.randn(ndim) for i in range(nwalkers)]

    if model == "sersic":
        func = lnprob_s

    elif model == "kazantzidis":
        func = lnprob_k

    elif model == "plummer":
        func = lnprob_p

    elif model == "gplummer":
        func = lnprob_gp

    if use_pool:

        with Pool() as pool:
            sampler = emcee.EnsembleSampler(
                nwalkers, ndim, func, args=(ri, ini, bounds), pool=pool
            )
            sampler.run_mcmc(pos, steps)
    else:

        sampler = emcee.EnsembleSampler(nwalkers, ndim, func, args=(ri, ini, bounds))
        sampler.run_mcmc(pos, steps)

    chain = sampler.chain

    return chain



"""
Created on 2020

@author: Eduardo Vitral
"""
###############################################################################
#
# 2020, Paris
#
# This file is based on the Python corner package (Copyright 2013-2016
# Dan Foreman-Mackey & contributors, The Journal of Open Source Software):
# https://joss.theoj.org/papers/10.21105/joss.00024
# I have done some modifications on it so it allows some new features and
# so it takes into account some choices as default. I thank Gary Mamon
# for his good suggestions concerning the plot visualization.
#
###############################################################################

from __future__ import print_function, absolute_import
import logging
import numpy as np
import matplotlib.pyplot as pl
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import LinearSegmentedColormap, colorConverter
from matplotlib.ticker import ScalarFormatter

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

try:
    from scipy.ndimage import gaussian_filter
except ImportError:
    gaussian_filter = None

__all__ = ["corner", "hist2d", "quantile"]


def dgauss(x, mu, sig, A):
    f = A * np.exp(-0.5 * ((x - mu) / sig) ** 2)
    return f


def corner(
    xs,
    range=None,
    weights=None,
    colormain="tab:blue",
    colorbackgd=(250 / 255, 245 / 255, 228 / 255),
    colorhist=(25 / 255, 70 / 255, 104 / 255),
    colorgauss="tab:green",
    labels=None,
    smooth=True,
    smooth1d=False,
    gauss_prior=None,
    gauss_mu=None,
    gauss_sig=None,
    truths=None,
    reals=None,
    truth_color="red",
    reals_color="orange",
    marker="X",
    own_dist=True,
    scale_hist=False,
    quantiles=None,
    verbose=False,
    fig=None,
    max_n_ticks=4,
    use_math_text=True,
    label_size=None,
    tick_size=None,
    dist_title=None,
    dist_ylab=None,
    tick_rotate=0,
    title_above=True,
    tick_above=True,
    prior_display=0,
):
    """
    Make a *sick* corner plot showing the projections of a data set in a
    multi-dimensional space. kwargs are passed to hist2d() or used for
    `matplotlib` styling.

    Parameters
    ----------
    xs : array_like[nsamples, ndim]
        The samples. This should be a 1- or 2-dimensional array. For a 1-D
        array this results in a simple histogram. For a 2-D array, the zeroth
        axis is the list of samples and the next axis are the dimensions of
        the space.

    range : float(ndim,2)
        Array containing the ranges to be used in the plots.

    weights : array_like[nsamples,]
        The weight of each sample. If `None` (default), samples are given
        equal weight.

    color : str
        A ``matplotlib`` style color for all histograms.

    labels : iterable (ndim,)
        A list of names for the dimensions. If a ``xs`` is a
        ``pandas.DataFrame``, labels will default to column names.

    smooth, smooth1d : float
       The standard deviation for Gaussian kernel passed to
       `scipy.ndimage.gaussian_filter` to smooth the 2-D and 1-D histograms
       respectively. If `None` (default), no smoothing is applied.

    gauss_prior : boolean array
        Boolean array with size equals the number of displayed parameters.
        It is True if the variable had Gaussian priors and False if not.
        It will display the relative box with green contours.

    gauss_mu : array
        Array containing the mean of the Gaussian priors.

    gauss_sig : array
        Array containing the dispersion of the Gaussian priors.

    truths : iterable (ndim,)
        A list of reference values to indicate on the plots.  Individual
        values can be omitted by using ``None``.

    reals : iterable (ndim,)
        A list of second reference values to indicate on the plots.  Individual
        values can be omitted by using ``None``.

    truth_color : str
        A ``matplotlib`` style color for the ``truths`` makers.

    reals_color : str
        A ``matplotlib`` style color for the ``reals`` makers.

    marker : str
        A ``matplotlib`` marker style. The default is "X".

    own_dist : Boolean
        True if the user wants to use specific implemantations for ticks/
        tiles/labels sizes. False if the user wants to use it as in the
        Python corner package.

    scale_hist : bool
        Should the 1-D histograms be scaled in such a way that the zero line
        is visible?

    quantiles : iterable
        A list of fractional quantiles to show on the 1-D histograms as
        vertical dashed lines.

    verbose : bool
        If true, print the values of the computed quantiles.

    max_n_ticks: int
        Maximum number of ticks to try to use

    fig : matplotlib.Figure
        Overplot onto the provided figure object.

    use_math_text : bool
        If true, then axis tick labels for very large or small exponents will
        be displayed as powers of 10 rather than using `e`.

    label_size : float
        Size of plot labels

    tick_size : float
        Size of plot ticks

    dist_title : float
        Distance between the title and the box

    dist_ylab : float
        Distance between the y-axis title and the box

    tick_rotate : float
        Rotation angle of x-ticks, in degrees.

    title_above : Boolean
        Boolean indicating it the user wants the label above the box.

    tick_above : Boolean
        Boolean indicating it the user wants the ticks above the box.

    prior_display : int
        Integer indicating the kind of display of the gaussian prior:
            0: No display: Default.
            1: Green box: Must provide "gauss_prior".
            2: Green Gaussian: Must provide valid "gauss_prior" and "range".

    """

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if quantiles is None:
        quantiles = []

    # Try filling in labels from pandas.DataFrame columns.
    if labels is None:
        try:
            labels = np.shape(xs)[1] * [""]
        except AttributeError:
            pass

    # Deal with 1D sample lists.
    xs = np.atleast_1d(xs)
    if len(xs.shape) == 1:
        xs = np.atleast_2d(xs)
    else:
        assert len(xs.shape) == 2, "The input sample array must be 1- or 2-D."
        xs = xs.T
    assert xs.shape[0] <= xs.shape[1], (
        "I don't believe that you want more " "dimensions than samples!"
    )

    # Parse the weight array.
    if weights is not None:
        weights = np.asarray(weights)
        if weights.ndim != 1:
            raise ValueError("Weights must be 1-D")
        if xs.shape[1] != weights.shape[0]:
            raise ValueError("Lengths of weights must match number of samples")

    range_provided = False
    # Parse the parameter ranges.
    if range is None:

        range = [[x.min(), x.max()] for x in xs]
        # Check for parameters that never change.
        m = np.array([e[0] == e[1] for e in range], dtype=bool)
        if np.any(m):
            raise ValueError(
                (
                    "It looks like the parameter(s) in "
                    "column(s) {0} have no dynamic range. "
                    "Please provide a `range` argument."
                ).format(", ".join(map("{0}".format, np.arange(len(m))[m])))
            )

    else:
        range_provided = True
        # If any of the extents are percentiles, convert them to ranges.
        # Also make sure it's a normal list.
        range = list(range)
        for i, _ in enumerate(range):
            try:
                emin, emax = range[i]
            except TypeError:
                q = [0.5 - 0.5 * range[i], 0.5 + 0.5 * range[i]]
                range[i] = quantile(xs[i], q, weights=weights)

    if len(range) != xs.shape[0]:
        raise ValueError("Dimension mismatch between samples and range")

    # Parse the bin specifications.
    bins = np.zeros(len(range)).astype(int)

    # Define labels size
    n_params = len(range)
    if n_params > 10:
        var = 10
    else:
        var = n_params

    if n_params > 6:
        max_n_ticks -= 1
    # if (n_params > 4) :
    #     max_n_ticks -= 1

    # Change here the sizes of labels/titles and ticks
    if label_size is None:
        label_size = var * 0.625 + 14
    if tick_size is None:
        tick_size = var * 0.5 + 13
    if dist_title is None:
        if tick_rotate == 0:
            dist_title = var * 0.008 + 1.15
        else:
            dist_title = (var * 0.008 + 1.15) * 1.1
    if dist_ylab is None:
        dist_ylab = var * 0.025 + 0.25

    # Some magic numbers for pretty axis layout.
    K = len(xs)
    factor = 2.0  # size of one side of one panel
    lbdim = 0.5 * factor  # size of left/bottom margin
    trdim = 0.2 * factor  # size of top/right margin
    whspace = 0  # w/hspace size
    plotdim = factor * K + factor * (K - 1.0) * whspace
    dim = lbdim + plotdim + trdim

    # Create a new figure if one wasn't provided.
    if fig is None:
        fig, axes = pl.subplots(K, K, figsize=(dim, dim))
    else:
        try:
            axes = np.array(fig.axes).reshape((K, K))
        except ValueError:
            raise ValueError(
                "Provided figure has {0} axes, but data has "
                "dimensions K={1}".format(len(fig.axes), K)
            )

    # Format the figure.
    lb = lbdim / dim
    tr = (lbdim + plotdim) / dim
    fig.subplots_adjust(
        left=lb, bottom=lb, right=tr, top=tr, wspace=whspace, hspace=whspace
    )

    for i, x in enumerate(xs):

        # Deal with masked arrays.
        if hasattr(x, "compressed"):
            x = x.compressed()

        if np.shape(xs)[0] == 1:
            ax = axes
        else:
            ax = axes[i, i]
        # Plot the histograms.

        # Define the bin quantity for each histogram
        q_16, q_50, q_84 = quantile(x, [0.16, 0.5, 0.84], weights=weights)
        q_m, q_p = q_50 - q_16, q_84 - q_50
        bins[i] = int((np.amax(range[i]) - np.amin(range[i])) / (min(q_m, q_p) / 4))

        if smooth1d:
            smooth1d = bins[i]

        if smooth1d is False:
            n, b, p = ax.hist(
                x,
                bins=bins[i],
                weights=weights,
                range=np.sort(range[i]),
                ec="k",
                histtype="stepfilled",
                color=colormain,
            )
            if gauss_prior is not None:
                if gauss_prior[i]:
                    if prior_display == 2:
                        if range_provided:
                            xx = np.linspace(min(range[i]), max(range[i]), 50)
                            disp0 = gauss_sig[i]
                            mu0 = gauss_mu[i]
                            amp = np.amax(n)
                            ax.plot(
                                xx,
                                dgauss(xx, mu0, disp0, amp),
                                color=colorgauss,
                                lw=4,
                                alpha=0.8,
                            )

            for axis in ["top", "bottom", "left", "right"]:
                ax.spines[axis].set_linewidth(2.2)
                if gauss_prior is not None:
                    if prior_display == 0:
                        prior_display = 1
                    if gauss_prior[i]:
                        if prior_display == 1:
                            ax.spines[axis].set_color(colorgauss)

        else:
            if gaussian_filter is None:
                raise ImportError("Please install scipy for smoothing")

            n, b, p = ax.hist(
                x,
                bins=bins[i],
                weights=weights,
                range=np.sort(range[i]),
                ec="k",
                histtype="stepfilled",
                color=colormain,
            )
            if gauss_prior is not None:
                if gauss_prior[i]:
                    if prior_display == 2:
                        if range_provided:
                            xx = np.linspace(min(range[i]), max(range[i]), 50)
                            disp0 = gauss_sig[i]
                            mu0 = gauss_mu[i]
                            amp = np.amax(n)
                            ax.plot(
                                xx,
                                dgauss(xx, mu0, disp0, amp),
                                color=colorgauss,
                                lw=4,
                                alpha=0.8,
                            )

            for axis in ["top", "bottom", "left", "right"]:
                ax.spines[axis].set_linewidth(2.2)
                if gauss_prior is not None:
                    if prior_display == 0:
                        prior_display = 1
                    if gauss_prior[i]:
                        if prior_display == 1:
                            ax.spines[axis].set_color(colorgauss)

        if truths is not None and truths[i] is not None:

            arrow_width = (np.amax(b) - np.amin(b)) / 20
            arrow_length = (np.amax(n) - np.amin(n)) / 12

            ax.arrow(
                truths[i],
                0,
                0,
                np.amax(n) * 0.6,
                head_width=arrow_width,
                head_length=arrow_length,
                color=truth_color,
                width=arrow_width / 4,
                zorder=10,
            )

        if reals is not None and reals[i] is not None:

            arrow_width = (np.amax(b) - np.amin(b)) / 20
            arrow_length = (np.amax(n) - np.amin(n)) / 12

            ax.arrow(
                reals[i],
                0,
                0,
                np.amax(n) * 0.4,
                head_width=arrow_width,
                head_length=arrow_length,
                color=reals_color,
                width=arrow_width / 4,
                zorder=10,
            )

        # Add in the column name if it's given.
        if labels is not None:
            title = "{0}".format(labels[i])
        else:
            raise ValueError("You did not give labels for your parameters.")

        if title is not None:
            if title_above:
                if own_dist:
                    ax.set_title(title, y=dist_title, fontsize=label_size)
                else:
                    ax.set_title(title)
        # Set up the axes.
        ax.set_xlim(range[i])
        ax.set_ylim(0, 1.1 * np.max(n))
        ax.set_yticklabels([])
        ax.xaxis.set_major_locator(MaxNLocator(max_n_ticks, prune="lower"))
        ax.yaxis.set_major_locator(MaxNLocator(max_n_ticks, prune="lower"))

        ax.tick_params(
            labeltop=tick_above,
            labelright=False,
            top=True,
            right=True,
            axis="both",
            which="major",
            direction="in",
            length=8,
            width=1.3,
            labelsize=tick_size,
        )
        ax.tick_params(
            labeltop=tick_above,
            labelright=False,
            top=True,
            right=True,
            axis="both",
            which="minor",
            direction="in",
            length=4,
            width=1.3,
            labelsize=tick_size,
        )
        if tick_rotate != 0:
            ax.tick_params(axis="x", labelrotation=tick_rotate)
        ax.minorticks_on()

        # Take off ticks that overlap left neighbour
        locations = ax.xaxis.get_ticklocs()
        size_x = max(range[i]) - min(range[i])
        if np.amin(locations) - min(range[i]) < size_x / 9:
            ax.xaxis.get_major_ticks()[0].label1.set_visible(False)
        if np.amin(locations) - min(range[i]) < size_x / 10:
            ax.xaxis.get_major_ticks()[0].label2.set_visible(False)

        if i < K - 1:
            pass

        else:
            [labs.set_rotation(tick_rotate) for labs in ax.get_xticklabels()]
            if labels is not None:
                if own_dist:
                    ax.set_xlabel(labels[i], fontsize=label_size)
                else:
                    ax.set_xlabel(labels[i])
                if tick_rotate != 0:
                    ax.xaxis.set_label_coords(0.5, -dist_ylab)
                else:
                    ax.xaxis.set_label_coords(0.5, -0.2)

        # use MathText for axes ticks
        ax.xaxis.set_major_formatter(
            ScalarFormatter(useMathText=use_math_text, useOffset=False)
        )

        for j, y in enumerate(xs):
            if np.shape(xs)[0] == 1:
                ax = axes
            else:
                ax = axes[i, j]
            if j > i:
                ax.set_frame_on(False)
                ax.set_xticks([])
                ax.set_yticks([])
                continue
            elif j == i:
                continue

            # Deal with masked arrays.
            if hasattr(y, "compressed"):
                y = y.compressed()

            hist2d(
                y,
                x,
                ax=ax,
                range=[range[j], range[i]],
                weights=weights,
                ccolormain=colormain,
                ccolorbackgd=colorbackgd,
                ccolorhist=colorhist,
                bins=[bins[j], bins[i]],
                smooth=smooth,
                ticksize=tick_size,
            )

            if truths is not None:
                if truths[i] is not None and truths[j] is not None:
                    # "What's at the X? Pirate treasure?"
                    ax.plot(
                        truths[j], truths[i], marker, color=truth_color, markersize=10
                    )

            if reals is not None:
                if reals[i] is not None and reals[j] is not None:
                    # "What's at the X? Pirate treasure?"
                    ax.plot(reals[j], reals[i], marker, color=reals_color, markersize=7)

            ax.xaxis.set_major_locator(MaxNLocator(max_n_ticks, prune="lower"))
            ax.yaxis.set_major_locator(MaxNLocator(max_n_ticks, prune="lower"))
            if i < K - 1:
                ax.set_xticklabels([])
            else:
                [labs.set_rotation(tick_rotate) for labs in ax.get_xticklabels()]
                if labels is not None:

                    # This concerns the titles on the bottom of the corner plot
                    if own_dist:
                        ax.set_xlabel(labels[j], fontsize=label_size)
                    if tick_rotate != 0:
                        ax.xaxis.set_label_coords(0.5, -dist_ylab)
                    else:
                        ax.xaxis.set_label_coords(0.5, -0.2)

                # Take off ticks that overlap left neighbour
                locations = ax.xaxis.get_ticklocs()
                size_x = max(range[j]) - min(range[j])
                if np.amin(locations) - min(range[j]) < size_x / 10:
                    ax.xaxis.get_major_ticks()[0].label1.set_visible(False)
                if max(range[j] - np.amax(locations)) < size_x / 10:
                    arg = len(locations) - 1
                    if max(range[j]) - locations[arg] < size_x / 10:
                        ax.xaxis.get_major_ticks()[arg].label1.set_visible(False)

            if j > 0:
                ax.set_yticklabels([])
            else:
                [labs.set_rotation(tick_rotate) for labs in ax.get_yticklabels()]
                if labels is not None:
                    # This concerns the titles in the left side of the corner
                    # plot
                    if own_dist:
                        ax.set_ylabel(labels[i], fontsize=label_size)
                        ax.yaxis.set_label_coords(-dist_ylab, 0.5)
                    if tick_rotate != 0:
                        ax.xaxis.set_label_coords(0.5, -dist_ylab)
                    else:
                        ax.xaxis.set_label_coords(0.5, -0.2)

                # Take off ticks that overlap bottom neighbour
                locations = ax.yaxis.get_ticklocs()
                size_x = max(range[i]) - min(range[i])
                if np.amin(locations) - min(range[i]) < size_x / 10:
                    ax.yaxis.get_major_ticks()[0].label1.set_visible(False)

    return fig


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


def quantile(x, q, weights=None):
    """
    Compute sample quantiles with support for weighted samples.

    Note
    ----
    When ``weights`` is ``None``, this method simply calls numpy's percentile
    function with the values of ``q`` multiplied by 100.

    Parameters
    ----------
    x : array_like[nsamples,]
       The samples.

    q : array_like[nquantiles,]
       The list of quantiles to compute. These should all be in the range
       ``[0, 1]``.

    weights : Optional[array_like[nsamples,]]
        An optional weight corresponding to each sample. These

    Returns
    -------
    quantiles : array_like[nquantiles,]
        The sample quantiles computed at ``q``.

    Raises
    ------
    ValueError
        For invalid quantiles; ``q`` not in ``[0, 1]`` or dimension mismatch
        between ``x`` and ``weights``.

    """
    x = np.atleast_1d(x)
    q = np.atleast_1d(q)

    if np.any(q < 0.0) or np.any(q > 1.0):
        raise ValueError("Quantiles must be between 0 and 1")

    if weights is None:
        return np.percentile(x, 100.0 * q)
    else:
        weights = np.atleast_1d(weights)
        if len(x) != len(weights):
            raise ValueError("Dimension mismatch: len(weights) != len(x)")
        idx = np.argsort(x)
        sw = weights[idx]
        cdf = np.cumsum(sw)[:-1]
        cdf /= cdf[-1]
        cdf = np.append(0, cdf)
        return np.interp(q, cdf, x[idx]).tolist()


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


def hist2d(
    x,
    y,
    bins=None,
    ticksize=5,
    range=None,
    weights=None,
    levels=None,
    smooth=True,
    ax=None,
    ccolormain=None,
    ccolorbackgd=None,
    ccolorhist=None,
    plot_datapoints=True,
    plot_density=True,
    plot_contours=True,
    no_fill_contours=False,
    fill_contours=True,
):
    """
    Plot a 2-D histogram of samples.

    Parameters
    ----------
    x : array_like[nsamples,]
       The samples.

    y : array_like[nsamples,]
       The samples.

    levels : array_like
        The contour levels to draw.

    ax : matplotlib.Axes
        A axes instance on which to add the 2-D histogram.

    plot_datapoints : bool
        Draw the individual data points.

    plot_density : bool
        Draw the density colormap.

    plot_contours : bool
        Draw the contours.

    no_fill_contours : bool
        Add no filling at all to the contours (unlike setting
        ``fill_contours=False``, which still adds a white fill at the densest
        points).

    fill_contours : bool
        Fill the contours.

    contour_kwargs : dict
        Any additional keyword arguments to pass to the `contour` method.

    contourf_kwargs : dict
        Any additional keyword arguments to pass to the `contourf` method.

    data_kwargs : dict
        Any additional keyword arguments to pass to the `plot` method when
        adding the individual data points.

    """
    if ax is None:
        ax = pl.gca()

    ax.tick_params(
        labeltop=False,
        labelright=False,
        top=True,
        right=True,
        axis="both",
        which="major",
        direction="in",
        length=8,
        width=1.3,
        labelsize=ticksize,
    )
    ax.tick_params(
        labeltop=False,
        labelright=False,
        top=True,
        right=True,
        axis="both",
        which="minor",
        direction="in",
        length=4,
        width=1.3,
        labelsize=ticksize,
    )
    ax.minorticks_on()

    ax.set_facecolor(ccolorbackgd)

    # Set the default range based on the data range if not provided.
    if range is None:
        range = [[x.min(), x.max()], [y.min(), y.max()]]

    # Set up the default plotting arguments.
    if ccolormain is None:
        ccolormain = "k"

    # Choose the default "sigma" contour levels.
    if levels is None:
        levels = 1.0 - np.exp(-0.5 * np.array([1, 2, 3]) ** 2)

    # This color map is used to hide the points at the high density areas.
    white_cmap = LinearSegmentedColormap.from_list(
        "white_cmap", [(1, 1, 1), (1, 1, 1)], N=2
    )

    # This "color map" is the list of colors for the contour levels if the
    # contours are filled.
    color_hist = ccolorhist
    rgba_color = colorConverter.to_rgba(color_hist)
    contour_cmap = [list(rgba_color) for levs in levels] + [rgba_color]
    for i, levs in enumerate(levels):
        contour_cmap[i][-1] *= float(i) / (len(levels) + 1)

    # We'll make the 2D histogram to directly estimate the density.
    try:
        H, X, Y = np.histogram2d(
            x.flatten(),
            y.flatten(),
            bins=bins,
            range=list(map(np.sort, range)),
            weights=weights,
        )
    except ValueError:
        raise ValueError(
            "It looks like at least one of your sample columns "
            "have no dynamic range. You could try using the "
            "'range' argument."
        )

    if smooth:

        factor = 5
        sig_smooth = [factor * np.std(x.flatten()), factor * np.std(y.flatten())]

        if gaussian_filter is None:
            raise ImportError("Please install scipy for smoothing")
        H = gaussian_filter(H, sig_smooth)

    # Compute the density levels.
    Hflat = H.flatten()
    inds = np.argsort(Hflat)[::-1]
    Hflat = Hflat[inds]
    sm = np.cumsum(Hflat)
    sm /= sm[-1]
    V = np.empty(len(levels))
    for i, v0 in enumerate(levels):
        try:
            V[i] = Hflat[sm <= v0][-1]
        except ImportError:
            V[i] = Hflat[0]
    V.sort()
    m = np.diff(V) == 0
    if np.any(m):
        logging.warning("Too few points to create valid contours")
    while np.any(m):
        V[np.where(m)[0][0]] *= 1.0 - 1e-4
        m = np.diff(V) == 0
    V.sort()

    # Compute the bin centers.
    X1, Y1 = 0.5 * (X[1:] + X[:-1]), 0.5 * (Y[1:] + Y[:-1])

    # Extend the array for the sake of the contours at the plot edges.
    H2 = H.min() + np.zeros((H.shape[0] + 4, H.shape[1] + 4))
    H2[2:-2, 2:-2] = H
    H2[2:-2, 1] = H[:, 0]
    H2[2:-2, -2] = H[:, -1]
    H2[1, 2:-2] = H[0]
    H2[-2, 2:-2] = H[-1]
    H2[1, 1] = H[0, 0]
    H2[1, -2] = H[0, -1]
    H2[-2, 1] = H[-1, 0]
    H2[-2, -2] = H[-1, -1]
    X2 = np.concatenate(
        [
            X1[0] + np.array([-2, -1]) * np.diff(X1[:2]),
            X1,
            X1[-1] + np.array([1, 2]) * np.diff(X1[-2:]),
        ]
    )
    Y2 = np.concatenate(
        [
            Y1[0] + np.array([-2, -1]) * np.diff(Y1[:2]),
            Y1,
            Y1[-1] + np.array([1, 2]) * np.diff(Y1[-2:]),
        ]
    )

    # Plot the base fill to hide the densest data points.
    if (plot_contours or plot_density) and not no_fill_contours:
        ax.contourf(
            X2, Y2, H2.T, [V.min(), H.max()], cmap=white_cmap, antialiased=False
        )

    if plot_contours and fill_contours:
        contourf_kwargs = dict()
        contourf_kwargs["colors"] = contourf_kwargs.get("colors", contour_cmap)
        contourf_kwargs["antialiased"] = contourf_kwargs.get("antialiased", False)
        ax.contourf(
            X2,
            Y2,
            H2.T,
            np.concatenate([[0], V, [H.max() * (1 + 1e-4)]]),
            **contourf_kwargs
        )

    # Plot the contour edge colors.
    if plot_contours:
        contour_kwargs = dict()
        contour_kwargs["colors"] = contour_kwargs.get("colors", color_hist)
        ax.contour(X2, Y2, H2.T, V, **contour_kwargs)
    for axis in ["top", "bottom", "left", "right"]:
        ax.spines[axis].set_linewidth(2.2)

    ax.set_xlim(range[0])
    ax.set_ylim(range[1])



"""
Created on 2021

@author: Eduardo Vitral
"""

###############################################################################
#
# November 2021, Paris
#
# This file contains the main functions concerning the creation and handling
# of mock data sets.
#
# Documentation is provided on Vitral, 2021.
# If you have any further questions please email vitral@iap.fr
#
###############################################################################

from . import angle
import numpy as np
import os
import glob

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Writing files"
# ---------------------------------------------------------------------------


def read_columns(file, transp=False):
    """
    Read txt file.

    Parameters
    ----------
    file : string
        Path to file.
    transp : boolean, optional
        True if you want the transposed matrix of the file.
        Default is False

    Returns
    -------
    data : TYPE
        DESCRIPTION.

    """

    data = np.loadtxt(file)

    if transp is False:
        return data
    else:
        return data.T


def write_cart(data, path):
    """
    Writes a file of cartesian coordinates.

    Parameters
    ----------
    data : array_like
        Data to be written.
    path : string
        location to store the data.

    Raises
    ------
    ValueError
        Whenever the path already exists or the data shape is not good.

    Returns
    -------
    None.

    """

    os.chdir(path)
    for file in glob.glob("*.dat"):
        if file == path:
            raise ValueError("File with the same name already exists.")

    # Creates file
    astrometric = open(path, "w")

    if np.shape(data[0]) == 6:
        astrometric.write("# x y z v_x v_y v_z \n")
        astrometric.write("# kpc kpc kpc km/s km/s km/s \n")
        for i in range(0, np.shape(data)[1]):
            # write line to output file
            string = "{: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} "
            line = string.format(
                data[0, i], data[1, i], data[2, i], data[3, i], data[4, i], data[5, i]
            )
            astrometric.write(line)
            astrometric.write("\n")
            astrometric.close()

        astrometric.close()
    elif np.shape(data[0]) == 9:
        astrometric.write("# x y z v_x ev_x v_y ev_y v_z ev_z \n")
        astrometric.write("# kpc kpc kpc km/s km/s km/s km/s km/s km/s \n")
        for i in range(0, np.shape(data)[1]):
            # write line to output file
            string = (
                "{: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} "
                + " {: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} "
            )
            line = string.format(
                data[0, i],
                data[1, i],
                data[2, i],
                data[3, i],
                data[4, i],
                data[5, i],
                data[6, i],
                data[7, i],
                data[8, i],
            )
            astrometric.write(line)
            astrometric.write("\n")
            astrometric.close()
    else:
        astrometric.close()
        raise ValueError("Does not recognize data shape.")


def write_gaia(data, path):
    """
    Writes a file of sky coordinates.

    Parameters
    ----------
    data : array_like
        Data to be written.
    path : string
        location to store the data.

    Raises
    ------
    ValueError
        Whenever the path already exists or the data shape is not good.

    Returns
    -------
    None.

    """

    # Creates file
    astrometric = open(path, "w")

    if np.shape(data)[0] == 6:
        astrometric.write("# RA Dec D PMRA PMDec vLOS \n")
        for i in range(0, np.shape(data)[1]):
            # write line to output file
            string = "{: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} "
            line = string.format(
                data[0, i], data[1, i], data[2, i], data[3, i], data[4, i], data[5, i]
            )
            astrometric.write(line)
            astrometric.write("\n")

        astrometric.close()
    elif np.shape(data)[0] == 9:
        astrometric.write("# RA Dec D PMRA ePMRA PMDec ePMDec vLOS evLOS \n")
        for i in range(0, np.shape(data)[1]):
            # write line to output file
            string = (
                "{: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} "
                + " {: 9.4f} {: 9.4f} {: 9.4f} {: 9.4f} "
            )
            line = string.format(
                data[0, i],
                data[1, i],
                data[2, i],
                data[3, i],
                data[4, i],
                data[5, i],
                data[6, i],
                data[7, i],
                data[8, i],
            )
            astrometric.write(line)
            astrometric.write("\n")
        astrometric.close()
    else:
        astrometric.close()
        raise ValueError("Does not recognize data shape.")


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Random generators"
# ---------------------------------------------------------------------------


def sphcap_generator(a0, d0, rmax, n):
    """
    Generates an spherical cap uniformally distributed.

    Parameters
    ----------
    a0 : float
        Right ascention from origin.
    d0 : float
        Declination from origin.
    rmax : float
        Maximum radius from center in degrees.
    n : int
        length of data set.

    Returns
    -------
    a : array_like
        Right ascention in degrees.
    d : array_like
        Declination in degrees.

    """

    u_1 = np.random.uniform(0, 1, n)
    u_2 = np.random.uniform(0, 1, n)
    r = np.arccos((1 - u_1) + u_1 * np.cos(rmax * np.pi / 180))
    phi = 2 * np.pi * u_2

    a = np.zeros(n)
    d = np.zeros(n)

    for i in range(n):

        a[i], d[i] = angle.polar_to_sky(
            r[i], phi[i], a0 * np.pi / 180, d0 * np.pi / 180
        )

    return a, d


def pm_generator(a, d, n, pmra0=0, pmdec0=0):
    """
    Random field stars proper motions following a Pearson VII distribution.

    Parameters
    ----------
    a : float
        Proper motion scale radius.
    d : float
        Proper motion slope.
    n : int
        length of data set.
    pmra0 : float, optional
        Mean pmra of interlopers.
    pmdec0 : float, optional
        Mean pmra of interlopers.

    Returns
    -------
    pmra : array_like
        Random pmra variables.
    pmdec : array_like
        Random pmdec variables.

    """

    u_1 = np.random.uniform(0, 1, n)
    u_2 = np.random.uniform(0, 1, n)
    r = a * np.sqrt(u_1 ** (1 / (1 + 0.5 * d)) - 1.0)
    phi = 2 * np.pi * u_2

    pmra = r * np.cos(phi) + pmra0
    pmdec = r * np.sin(phi) + pmdec0

    return pmra, pmdec


def gaussian_generator(mean, dev):
    """
    Random Gaussian variables of mean "mean" and standard deviation "dev".

    Parameters
    ----------
    mean : array_like
        Gaussian mean.
    dev : array_like
        Gaussian dispersion.

    Returns
    -------
    x : array_like
        Random Gaussian variables.

    """

    x = np.random.normal(mean, dev, size=len(mean))

    return x


def magnitude_generator(n, a=0.25, b=21):
    """
    Random G magnitude generator. Supposes a Gaia like distribution.

    Parameters
    ----------
    n : int
        number of points to be generated.
    a : float
        slope of log10 magnitude cumulative distribution.
        The default is 0.25
    b : float
        Cutting G magnitude.
        The default is 21.

    Returns
    -------
    x : array_like
        Random magnitudes.

    """

    u = np.random.uniform(low=0.0, high=1.0, size=n)

    x = np.log10(u) / a + b

    return x


def get_error_pm(n):
    """
    Random proper motion error generator. Supposes a Gaia like distribution.

    Parameters
    ----------
    n : int
        number of points to be generated.

    Returns
    -------
    x : array_like
        Random error in pmra.
    y : array_like
        Random error in pmdec.

    """

    m = magnitude_generator(n)

    x = 10 ** (0.26 * (m - 21.5))
    y = 10 ** (0.26 * (m - 21.7))

    return x, y


def get_error_rv(n):
    """
    Random radial velocity generator. Supposes a Gaia like distribution.

    Parameters
    ----------
    n : int
        number of points to be generated.

    Returns
    -------
    x : array_like
        Random error in radial velocity.

    """

    m = magnitude_generator(n)

    x = 10 ** (0.0026 * m * m * m - 0.061 * m * m + 0.56 * m - 2.7)

    return x


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Data conversion"
# ---------------------------------------------------------------------------


def cart6d_to_gaia(data, error=False):
    """
    Transforms a cartesian 6D file into Gaia-like coordinates.

    Parameters
    ----------
    data : array_like
        Cartesian data.
    error : boolean, optional
        True if the user wants to provide Gaia like velocity errors.
        The default is False.

    Returns
    -------
    data_gaia : array_like
        Original file in Gaia-like coordinates.

    """

    coord = angle.cart_to_radec(data[0], data[1], data[2], data[3], data[4], data[5])

    RA = coord[0]
    Dec = coord[1]
    DD = coord[2]
    PMRA = coord[3]
    PMDec = coord[4]
    v_LOS = coord[5]

    if error is False:
        data_gaia = np.asarray([RA, Dec, DD, PMRA, PMDec, v_LOS])
    else:
        ePMRA, ePMDec = get_error_pm(len(PMRA))
        ev_LOS = get_error_rv(len(v_LOS))

        PMRA = gaussian_generator(PMRA, ePMRA)
        PMDec = gaussian_generator(PMDec, ePMDec)
        v_LOS = gaussian_generator(v_LOS, ev_LOS)

        data_gaia = np.asarray([RA, Dec, DD, PMRA, ePMRA, PMDec, ePMDec, v_LOS, ev_LOS])

    return data_gaia.astype(float)



"""
Created on 2020

@author: Eduardo Vitral
"""

###############################################################################
#
# November 2020, Paris
#
# This file contains the main functions concerning parallax information.
# It provides a kernel density estimation of the distance distribution,
# as well as a fit of the mode of this distribution.
#
# Documentation is provided on Vitral, 2021.
# If you have any further questions please email vitral@iap.fr
#
###############################################################################

import numpy as np
from scipy.optimize import differential_evolution
from scipy.stats import gaussian_kde
import numdifftools as ndt

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Mode estimation"
# ---------------------------------------------------------------------------


def kde(data, bw=None):
    """
    Computes the Kernel Density Estimation of the data.

    Parameters
    ----------
    data : array_like
        Data to be analyzed.
    bw : float, optional
        bandwidth method from the kde routine from scipy. The default is None.

    Returns
    -------
    kde object
        kde object of the data, using a specific bandwith.

    """

    if bw is None:
        bw = 0.5 * (len(data) * (1 + 2) / 4) ** (-1 / (1 + 4))

    return gaussian_kde(data, bw_method=bw)


def refined_mode(data, bw=None, use_log=False):
    """
    Computes the mode of data, by a KDE non-parametric fit.

    Parameters
    ----------
    data : array_like
        Data to be analyzed.
    bw : float, optional
        bandwidth method from the kde routine from scipy. The default is None.
    use_log : boolean, optional
        True is the user wants to find the mode on a logarithmic spaced data.
        The default is False.

    Returns
    -------
    mode : float
        Mode of the analyzed distribution.
    uncertainty : float
        Uncertainty on the mode measurement.

    """

    if bw is None:
        bw = 0.5 * (len(data) * (1 + 2) / 4) ** (-1 / (1 + 4))

    if use_log is False:
        mode, uncertainty = mode_lin(data, bw)
    else:
        mode, uncertainty = mode_log(data, bw)

    return mode, uncertainty


def mode_lin(data, bw):
    """
    Computes the mode from linear spaced data.

    Parameters
    ----------
    data : array_like
        Data to be analyzed.
    bw : float
        bandwidth method from the kde routine from scipy.

    Returns
    -------
    mode : float
        Mode of the analyzed distribution.
    uncertainty : float
        Uncertainty on the mode measurement.

    """

    kernel = kde(data, bw=bw)
    pdf = kernel.pdf(data)
    x0 = data[np.argmax(pdf)]
    span = data.max() - data.min()
    dx = span / 10
    bounds = np.array([[x0 - dx, x0 + dx]])

    results = differential_evolution(lambda x: -kernel(x), bounds)

    hfun = ndt.Hessian(lambda x: -kernel(x), full_output=True)
    hessian_ndt, info = hfun(results.x[0])
    var = np.sqrt(np.diag(np.linalg.inv(hessian_ndt)))

    return results.x[0], var[0]


def mode_log(data, bw):
    """
    Computes the mode from logarithmic spaced data.

    Parameters
    ----------
    data : array_like
        Data to be analyzed.
    bw : float
        bandwidth method from the kde routine from scipy.

    Returns
    -------
    mode : float
        Mode of the analyzed distribution.
    uncertainty : 2D array
        Lower and upper uncertainties on the mode measurement.

    """

    logdata = np.log(data)
    kernel = kde(logdata, bw=bw)
    height = kernel.pdf(logdata)
    x0 = logdata[np.argmax(height)]
    span = logdata.max() - logdata.min()
    dx = span / 10
    bounds = np.array([[x0 - dx, x0 + dx]])

    results = differential_evolution(lambda x: -kernel(x), bounds)

    hfun = ndt.Hessian(lambda x: -kernel(x), full_output=True)
    hessian_ndt, info = hfun(results.x[0])
    var = np.sqrt(np.diag(np.linalg.inv(hessian_ndt)))

    mode = np.exp(results.x[0])
    unc1 = np.exp(results.x[0]) - np.exp(results.x[0] - var[0])
    unc2 = np.exp(results.x[0] + var[0]) - np.exp(results.x[0])

    uncertainty = np.asarray([unc1, unc2])

    return mode, uncertainty


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"General functions"
# ---------------------------------------------------------------------------


def parallax_to_distance(par, remove_far=True):
    """
    Converts parallax into distance [kpc].

    Parameters
    ----------
    par : array_like
        Array containing parallaxes.
    remove_far : boolean, optional
        Set True is the user want to remove upper outliers.
        The default is True.

    Returns
    -------
    dd : array_like
        Distance in kpc.

    """

    dd = 1 / par

    dd = dd[np.where(dd > 0)]

    if remove_far is True:
        q975 = quantile(dd, 0.975)
        idx_far = np.where(dd < q975)
        dd = dd[idx_far]

    return dd


def quantile(x, q):
    """
    Compute sample quantiles.

    Parameters
    ----------
    x : array_like
        Array containing set of values.
    q : array_like, float
        Quantile values to be derived (must be between 0 and 1).

    Raises
    ------
    ValueError
        Quantiles must be between 0 and 1.

    Returns
    -------
    percentile: array_like (same shape as q), float
        Percentile value(s).

    """

    x = np.atleast_1d(x)
    q = np.atleast_1d(q)

    if np.any(q < 0.0) or np.any(q > 1.0):
        raise ValueError("Quantiles must be between 0 and 1")

    return np.percentile(x, 100.0 * q)



"""
Created on 2020

@author: Eduardo Vitral
"""

###############################################################################
#
# November 2020, Paris
#
# This file contains the main functions concerning the angular tansformations,
# sky projections and spherical trigonomtry.
#
# Documentation is provided on Vitral, 2021.
# If you have any further questions please email vitral@iap.fr
#
###############################################################################

import numpy as np

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Global variables"
# ------------------------------------------------------------------------------

# Right ascention of the north galactic pole, in radians
a_NGP = 192.85947789 * (np.pi / 180)
# Declination of the north galactic pole, in radians
d_NGP = 27.12825241 * (np.pi / 180)
# Longitude of the north celestial pole, in radians
l_NCP = 122.93192526 * (np.pi / 180)

# Vertical waves in the solar neighbourhood in Gaia DR2
# Bennett & Bovy, 2019, MNRAS
# --> Sun Z position (kpc), in galactocentric coordinates
z_sun = 0.0208

# Sun Y position (kpc), in galactocentric coordinates, by definition.
y_sun = 0

# A geometric distance measurement to the Galactic center black hole
# with 0.3% uncertainty
# Gracity Collaboration, 2019, A&A
# --> Sun distance from the Galactic center, in kpc.
d_sun = 8.178

# Sun X position (kpc), in galactocentric coordinates
x_sun = np.sqrt(d_sun * d_sun - z_sun * z_sun)

# On the Solar Velocity
# Ronald Drimmel and Eloisa Poggio, 2018, RNAAS
# --> Sun X velocity (km/s), in galactocentric coordinates
vx_sun = -12.9
# --> Sun Y velocity (km/s), in galactocentric coordinates
vy_sun = 245.6
# --> Sun Z velocity (km/s), in galactocentric coordinates
vz_sun = 7.78

# Multuplying factor to pass from kpc to km
kpc_to_km = 3.086 * 10 ** 16

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Angles handling"
# ------------------------------------------------------------------------------


def sky_distance_deg(RA, Dec, RA0, Dec0):
    """
    Computes the sky distance (in degrees) between two sets of
    sky coordinates, given also in degrees.

    Parameters
    ----------
    RA : array_like, float
        Right ascension (in degrees) of object 1.
    Dec : array_like (same shape as RA), float
        Declination (in degrees) of object 1.
    RA0 : array_like (same shape as RA), float
        Right ascension (in degrees) of object 2.
    Dec0 : array_like (same shape as RA), float
        Declination (in degrees) of object 2.

    Returns
    -------
    R : array_like, float
        Sky distance (in degrees) between object 1 and object 2.

    """

    RA = RA * np.pi / 180
    Dec = Dec * np.pi / 180

    RA0 = RA0 * np.pi / 180
    Dec0 = Dec0 * np.pi / 180

    R = (180 / np.pi) * np.arccos(
        np.sin(Dec) * np.sin(Dec0) + np.cos(Dec) * np.cos(Dec0) * np.cos((RA - RA0))
    )

    return np.asarray(R)


def get_circle_sph_trig(r, a0, d0, nbins=500):
    """
    Generates a circle in spherical coordinates.

    Parameters
    ----------
    r : float
        Distance from the center, in degrees.
    a0 : float
        Right ascention from origin, in degrees.
    d0 : float
        Declination from origin in, in degrees.
    nbins : int, optional
        Number of circle points. The default is 500.

    Returns
    -------
    ra : array_like
        Right ascention in degrees.
    dec : array_like
        Declination in degrees.

    """

    # Converts angles to radians
    r = r * np.pi / 180
    a0 = a0 * np.pi / 180
    d0 = d0 * np.pi / 180

    phi = np.linspace(0, 2 * np.pi, nbins)

    a = np.zeros(nbins)
    d = np.zeros(nbins)
    for i in range(0, nbins):
        a[i], d[i] = polar_to_sky(r, phi[i], a0, d0)

    return a, d


def polar_to_sky(r, phi, a0, d0):
    """
    Transforms spherical polar coordinates (r,phi) into sky coordinates,
    in degrees (RA,Dec).

    Parameters
    ----------
    r : array_like
        Radial distance from center.
    phi : array_like
        Angle between increasing declination and the projected radius
        (pointing towards the source).
    a0 : float
        Right ascention from origin in radians.
    d0 : float
        Declination from origin in radians.

    Returns
    -------
    ra : array_like
        Right ascention in degrees.
    dec : array_like
        Declination in degrees.

    """

    d = np.arcsin(np.cos(r) * np.sin(d0) + np.cos(d0) * np.cos(phi) * np.sin(r))

    if phi < np.pi:
        if (np.cos(r) - np.sin(d) * np.sin(d0)) / (np.cos(d) * np.cos(d0)) > 0:
            a = a0 + np.arccos(np.sqrt(1 - (np.sin(phi) * np.sin(r) / np.cos(d)) ** 2))
        else:
            a = a0 + np.arccos(-np.sqrt(1 - (np.sin(phi) * np.sin(r) / np.cos(d)) ** 2))

    if phi >= np.pi:
        if (np.cos(r) - np.sin(d) * np.sin(d0)) / (np.cos(d) * np.cos(d0)) > 0:
            a = a0 - np.arccos(np.sqrt(1 - (np.sin(phi) * np.sin(r) / np.cos(d)) ** 2))
        else:
            a = a0 - np.arccos(-np.sqrt(1 - (np.sin(phi) * np.sin(r) / np.cos(d)) ** 2))

    ra = a * 180 / np.pi
    dec = d * 180 / np.pi

    return ra, dec


def sky_to_polar(a, d, a0, d0):
    """
    Transforms sky coordinates, in degrees (RA,Dec), into
    spherical polar coordinates (r,phi).

    Parameters
    ----------
    a : array_like
        Right ascention in degrees.
    d : array_like
        Declination in degrees.
    a0 : float
        Right ascention from origin.
    d0 : float
        Declination from origin.

    Returns
    -------
    r : array_like
        Radial distance from center.
    p : array_like
        Angle between increasing declination and the projected radius
        (pointing towards the source), in radians.

    """

    r = sky_distance_deg(a, d, a0, d0) * np.pi / 180

    a = a * np.pi / 180
    d = d * np.pi / 180

    sp = np.cos(d) * np.sin(a - (a0 * np.pi / 180)) / np.sin(r)

    p = np.zeros(len(sp))

    spp = np.where(sp > 0)
    spm = np.where(sp <= 0)
    dp = np.where(d > (d0 * np.pi / 180))
    dm = np.where(d <= (d0 * np.pi / 180))

    p[np.intersect1d(spp, dp)] = np.arcsin(sp[np.intersect1d(spp, dp)])
    p[np.intersect1d(spp, dm)] = np.pi - np.arcsin(sp[np.intersect1d(spp, dm)])
    p[np.intersect1d(spm, dp)] = 2 * np.pi + np.arcsin(sp[np.intersect1d(spm, dp)])
    p[np.intersect1d(spm, dm)] = np.pi - np.arcsin(sp[np.intersect1d(spm, dm)])

    return r, p


def angular_sep_vector(v0, v):
    """
    Returns separation angle in radians between two 3D arrays.

    Parameters
    ----------
    v0 : 3D array
        Vector 1.
    v : 3D array
        Vector 2.

    Returns
    -------
    R : array_like
        Separation between vector 1 and vector 2 in radians.

    """

    try:
        v0.shape
    except NameError:
        print("You did not give a valid input.")
        return
    v = v / np.linalg.norm(v)
    B = np.arcsin(v[2])
    cosA = v[0] / np.cos(B)
    sinA = v[1] / np.cos(B)

    if cosA > 0 and sinA >= 0:
        A = np.arcsin(sinA)
    if cosA <= 0 and sinA > 0:
        A = np.arccos(cosA)
    if cosA < 0 and sinA <= 0:
        A = 2 * np.pi - np.arccos(cosA)
    if cosA >= 0 and sinA < 0:
        A = np.arcsin(sinA)

    v0 = v0 / np.linalg.norm(v0)
    B0 = np.arcsin(v0[2])
    cosA0 = v0[0] / np.cos(B0)
    sinA0 = v0[1] / np.cos(B0)

    if cosA0 > 0 and sinA0 >= 0:
        A0 = np.arcsin(sinA0)
    if cosA0 <= 0 and sinA0 > 0:
        A0 = np.arccos(cosA0)
    if cosA0 < 0 and sinA0 <= 0:
        A0 = 2 * np.pi - np.arccos(cosA0)
    if cosA0 >= 0 and sinA0 < 0:
        A0 = np.arcsin(sinA0)

    cosR = np.sin(B0) * np.sin(B) + np.cos(B0) * np.cos(B) * np.cos(A - A0)
    R = np.arccos(cosR)

    return R


def rodrigues_formula(k, v, theta, debug=False):
    """
    Returns the rotation of v of an angle theta with respect to the vector k
    Applies the Rodrigues formula from:
    https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula

    Parameters
    ----------
    k : array_like
        Vector with respect to which v will be rotated.
    v : array_like
        Vector to be rotated.
    theta : float
        Angle to rotate v.
    debug : boolean, optional
       True if the reader wants to print debug diagnistics.
       The default is False.

    Returns
    -------
    v_rot : array_like
        Rotated vector.

    """

    try:
        v.shape
    except NameError:
        print("You did not give a valid input for the Rodrigues formula.")
        return

    if len(np.shape(v)) == 1 and len(v) == 3:
        v_rot = (
            v * np.cos(theta)
            + np.cross(k, v) * np.sin(theta)
            + k * np.dot(k, v) * (1 - np.cos(theta))
        )

    elif len(np.shape(v)) == 2 and np.shape(v)[0] == 3:
        v_rot = np.zeros((np.shape(v)[1], 3))
        for i in range(0, len(v_rot)):

            v0 = np.asarray([v[0][i], v[1][i], v[2][i]])
            v_rot[i] = (
                v0 * np.cos(theta)
                + np.cross(k, v0) * np.sin(theta)
                + k * np.dot(v0, k) * (1 - np.cos(theta))
            )

            if debug is True and i < 10:
                print("v0   :", v0)
                print("v_rot:", v_rot[i])
    else:
        print("You did not give a valid input for the Rodrigues formula.")
        return

    return v_rot


def sky_coord_rotate(v_i, v0_i, v0_f, theta=0, debug=False):
    """
    Gets new angles (RA,Dec) in degrees of a rotated vector.

    Parameters
    ----------
    v_i : array_like
        Vectors to be rotated.
    v0_i : array_like
        Vector pointing to the initial centroid position.
    v0_f : array_like
        Vector pointing to the final centroid position.
    theta : float, optional
        Angle to rotate (for no translation), in radians. The default is 0.
    debug : boolean, optional
       True if the reader wants to print debug diagnistics.
       The default is False.

    Returns
    -------
    array_like, array_like
        Arrays containing the new rotated (RA,Dec) positions.

    """

    if (v0_i == v0_f).all():
        if debug is True:
            print("Pure rotation")
        k = v0_i / np.linalg.norm(v0_i)
    else:
        if debug is True:
            print("Translation in spherical geometry")
        k = np.cross(v0_i, v0_f) / np.linalg.norm(np.cross(v0_i, v0_f))
        theta = angular_sep_vector(v0_i, v0_f)

    if debug is True:
        print("Vector k:", k)
        print("Angle of separation [degrees]:", theta * 180 / np.pi)

    v_f = rodrigues_formula(k, v_i, theta)

    try:
        v_f.shape
    except NameError:
        print("You did not give a valid input for the Rodrigues formula.")
        return

    if len(np.shape(v_f)) == 1 and len(v_f) == 3:
        v_f = v_f / np.linalg.norm(v_f)
        B = np.arcsin(v_f[2])
        cosA = v_f[0] / np.cos(B)
        sinA = v_f[1] / np.cos(B)

        if cosA > 0 and sinA >= 0:
            A = np.arcsin(sinA)
        if cosA <= 0 and sinA > 0:
            A = np.arccos(cosA)
        if cosA < 0 and sinA <= 0:
            A = 2 * np.pi - np.arccos(cosA)
        if cosA >= 0 and sinA < 0:
            A = np.arcsin(sinA)

    elif len(np.shape(v_f)) == 2:
        A = np.zeros(len(v_f))
        B = np.zeros(len(v_f))
        for i in range(0, len(v_f)):
            v_f[i] = v_f[i] / np.linalg.norm(v_f[i])
            B[i] = np.arcsin(v_f[i][2])
            cosA = v_f[i][0] / np.cos(B[i])
            sinA = v_f[i][1] / np.cos(B[i])

            if cosA > 0 and sinA >= 0:
                A[i] = np.arcsin(sinA)
            if cosA <= 0 and sinA > 0:
                A[i] = np.arccos(cosA)
            if cosA < 0 and sinA <= 0:
                A[i] = 2 * np.pi - np.arccos(cosA)
            if cosA >= 0 and sinA < 0:
                A[i] = np.arcsin(sinA)

            if debug is True and i < 10:
                print("A, B [degrees]:", A[i] * (180 / np.pi), B[i] * (180 / np.pi))
    else:
        print("You did not give a valid input for the Rodrigues formula.")
        return

    return A * (180 / np.pi), B * (180 / np.pi)


def sky_vector(a, d, a0, d0, af, df):
    """
    Transforms sky coordinates in vectors to be used by sky_coord_rotate.

    Parameters
    ----------
    a : float, array_like
        Original set of RA in degrees.
    d : float, array_like
        Original set of Dec in degrees.
    a0 : float
        Original centroid RA in degrees.
    d0 : float
        Original centroid Dec in degrees.
    af : float
        Final centroid RA in degrees.
    df : float
        Final centroid Dec in degrees.

    Returns
    -------
    v_i : array_like
        Vectors to be rotated.
    v0_i : array_like
        Vector pointing to the initial centroid position.
    v0_f : array_like
        Vector pointing to the final centroid position.

    """

    a = a * (np.pi / 180)
    d = d * (np.pi / 180)
    a0 = a0 * (np.pi / 180)
    d0 = d0 * (np.pi / 180)
    af = af * (np.pi / 180)
    df = df * (np.pi / 180)

    v_i = np.asarray([np.cos(a) * np.cos(d), np.sin(a) * np.cos(d), np.sin(d)])

    v0_i = np.asarray([np.cos(a0) * np.cos(d0), np.sin(a0) * np.cos(d0), np.sin(d0)])

    v0_f = np.asarray([np.cos(af) * np.cos(df), np.sin(af) * np.cos(df), np.sin(df)])

    return v_i, v0_i, v0_f


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Axis rotation"
# ------------------------------------------------------------------------------


def transrot_source(a, d, a0, d0, af, df):
    """
    Translades and then rotates a source in spherical coordinates, so
    the original directions remain the same.

    Parameters
    ----------
    a : float, array_like
        Original set of RA in degrees.
    d : float, array_like
        Original set of Dec in degrees.
    a0 : float
        Original centroid RA in degrees.
    d0 : float
        Original centroid Dec in degrees.
    af : float
        Final centroid RA in degrees.
    df : float
        Final centroid Dec in degrees.

    Returns
    -------
    a : array_like
        Right ascention in degrees.
    d : array_like
        Declination in degrees.

    """

    v_i, v0_i, v0_f = sky_vector(a, d, a0, d0, af, df)

    a, d = sky_coord_rotate(v_i, v0_i, v0_f)

    rmax = np.nanmax(sky_distance_deg(a, d, af, df)) * (np.pi / 180)

    narr = np.asarray([0, 0.1, 0.25, 0.5]) * rmax
    vt = np.asarray(
        [
            np.cos(a0 + narr) * np.cos([d0, d0, d0, d0]),
            np.sin(a0 + narr) * np.cos([d0, d0, d0, d0]),
            np.sin([d0, d0, d0, d0]),
        ]
    )

    at, dt = sky_coord_rotate(vt, v0_i, v0_f)

    r0, p0 = sky_to_polar(a0 + narr, d0 * np.ones(len(narr)), a0, d0)
    rf, pf = sky_to_polar(at, dt, af, df)

    phi = np.nanmean(pf - p0)

    v_i, v0_i, v0_f = sky_vector(a, d, af, df, af, df)

    a, d = sky_coord_rotate(v_i, v0_i, v0_f, theta=phi)

    return a, d


def rotate_axis(x, y, theta, mu_x=0, mu_y=0):
    """
    Rotates and translates the two main cartesian axis.

    Parameters
    ----------
    x : float or array_like
        Data in x-direction.
    y : float or array_like
        Data in y-direction.
    theta : float
        Rotation angle in radians.
    mu_x : float, optional
        Center of new x axis in the old frame. The default is 0.
    mu_y : float, optional
        Center of new y axis in the old frame. The default is 0.

    Returns
    -------
    x_new : float or array_like
        Data in new x-direction.
    y_new : float or array_like
        Data in new y-direction.

    """

    x_new = (x - mu_x) * np.cos(theta) + (y - mu_y) * np.sin(theta)
    y_new = -(x - mu_x) * np.sin(theta) + (y - mu_y) * np.cos(theta)

    return x_new, y_new


def get_ellipse(a, b, theta, nbins):
    """
    Provides an array describing a rotated ellipse.

    Parameters
    ----------
    a : float
        Ellipse semi-major axis.
    b : float
        Ellipse semi-minor axis.
    theta : float
        Rotation angle in radians.
    nbins : int
        Number of points in the final array.

    Returns
    -------
    ellipse_rot : array_like
        Rotated ellipse array.

    """

    t = np.linspace(0, 2 * np.pi, nbins)
    ellipse = np.array([a * np.cos(t), b * np.sin(t)])
    m_rot = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
    ellipse_rot = np.zeros((2, ellipse.shape[1]))
    for i in range(ellipse.shape[1]):
        ellipse_rot[:, i] = np.dot(m_rot, ellipse[:, i])

    return ellipse_rot


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Transform coordinates"
# ------------------------------------------------------------------------------


def cart_to_sph(x, y, z, vx, vy, vz):
    """
    Transforms 6D cartesian coordinates to spherical coordinates.

    Parameters
    ----------
    x : array_like, float
        x-axis.
    y : array_like, float
        y-axis.
    z : array_like, float
        z-axis.
    vx : array_like, float
        x-axis velocity.
    vy : array_like, float
        y-axis velocity.
    vz : array_like, float
        z-axis velocity.

    Returns
    -------
    r : array_like, float
        r-axis.
    phi : array_like, float
        phi-angle.
    theta : array_like, float
        theta-angle.
    vr : array_like, float
        r-axis velocity.
    vphi : array_like, float
        phi-angle velocity.
    vtheta : array_like, float
        theta-angle velocity.

    """

    r = np.sqrt(x * x + y * y + z * z)
    phi = np.arctan(y / x)
    theta = np.arccos(z / r)

    vr = (vx * x + vy * y + vz * z) / r
    vphi = r * (vy * x - vx * y) / (x * x + y * y)
    vtheta = -(vz * (x * x + y * y) - z * (vx * x + vy * y)) / (
        np.sqrt(x * x + y * y) * r
    )

    return r, phi, theta, vr, vphi, vtheta


def sph_to_cart(r, phi, theta, vr, vphi, vtheta):
    """
    Transforms 6D spherical coordinates to cartesian coordinates.

    Parameters
    ----------
    r : array_like, float
        r-axis.
    phi : array_like, float
        phi-angle.
    theta : array_like, float
        theta-angle.
    vr : array_like, float
        r-axis velocity.
    vphi : array_like, float
        phi-angle velocity.
    vtheta : array_like, float
        theta-angle velocity.

    Returns
    -------
    x : array_like, float
        x-axis.
    y : array_like, float
        y-axis.
    z : array_like, float
        z-axis.
    vx : array_like, float
        x-axis velocity.
    vy : array_like, float
        y-axis velocity.
    vz : array_like, float
        z-axis velocity.

    """

    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)

    vx = (
        vr * np.sin(theta) * np.cos(phi)
        + vtheta * np.cos(theta) * np.cos(phi)
        - vphi * np.sin(theta) * np.sin(phi)
    )
    vy = (
        vr * np.sin(theta) * np.sin(phi)
        + vtheta * np.cos(theta) * np.sin(phi)
        + vphi * np.sin(theta) * np.cos(phi)
    )
    vz = vr * np.cos(theta) - vtheta * np.sin(theta)

    return x, y, z, vx, vy, vz


def radec_to_lb(a, d, dadt=None, dddt=None):
    """
    Transforms celestial coordinates into galactic coordinates.

    Parameters
    ----------
    a : array_like, float
        Right ascention in degrees.
    d : array_like, float
        Declination in degrees.
    dadt : array_like, float, optional
        Right ascention velocity (PMRA), in mas/yr.
        The default is None
    dddt : array_like, float, optional
        Declination velocity (PMDec), in mas/yr.
        The default is None

    Returns
    -------
    lon : array_like, float
        Galactic longitude, in degrees.
    b : array_like, float
        Galactic latitude, in degrees.
    dldt : array_like, float, optional
        Galactic longitude velocity, in mas/yr
    dbdt : array_like, float, optional
        Galactic latitude velocity, in mas/yr. The default is None

    """

    a = a * (np.pi / 180)
    d = d * (np.pi / 180)

    sind = np.sin(d)
    cosd = np.cos(d)
    sind_NGP = np.sin(d_NGP)
    cosd_NGP = np.cos(d_NGP)
    cosda = np.cos(a - a_NGP)
    sinda = np.sin(a - a_NGP)

    sinb = sind_NGP * sind + cosd_NGP * cosd * cosda
    cosb_sindl = cosd * sinda
    cosb_cosdl = cosd_NGP * sind - sind_NGP * cosd * cosda

    b = np.arcsin(sinb)

    cosb = np.cos(b)

    sindl = cosb_sindl / cosb
    cosdl = cosb_cosdl / cosb

    if np.isscalar(sindl):
        if (sindl >= 0 and cosdl > 0) or (sindl < 0 and cosdl >= 0):
            lon = l_NCP - np.arctan(cosb_sindl / cosb_cosdl)
        elif (sindl > 0 and cosdl <= 0) or (sindl <= 0 and cosdl < 0):
            lon = l_NCP - np.arctan(cosb_sindl / cosb_cosdl) + np.pi
    else:
        lon = np.zeros(len(sindl))
        for i in range(len(sindl)):
            if (sindl[i] >= 0 and cosdl[i] > 0) or (sindl[i] < 0 and cosdl[i] >= 0):
                lon[i] = l_NCP - np.arctan(cosb_sindl[i] / cosb_cosdl[i])
            elif (sindl[i] > 0 and cosdl[i] <= 0) or (sindl[i] <= 0 and cosdl[i] < 0):
                lon[i] = l_NCP - np.arctan(cosb_sindl[i] / cosb_cosdl[i]) + np.pi

    if np.isscalar(lon):
        lon = lon % (2 * np.pi)
    else:
        for i in range(len(lon)):
            lon[i] = lon[i] % (2 * np.pi)

    if dadt is not None and dddt is not None:
        sinb = np.sin(b)
        cosdl = np.cos(l_NCP - lon)

        dbdt = (
            sind_NGP * cosd * dddt
            - cosd_NGP * cosda * sind * dddt
            - cosd_NGP * sinda * dadt
        ) / cosb

        dldt = (
            cosb * (sind * dddt * sinda - cosda * dadt) - cosd * sinda * sinb * dbdt
        ) / (cosb * cosb * cosdl)

        lon = lon * (180 / np.pi)
        b = b * (180 / np.pi)

        return lon, b, dldt, dbdt

    else:

        lon = lon * (180 / np.pi)
        b = b * (180 / np.pi)

        return lon, b


def lb_to_radec(lon, b, dldt=None, dbdt=None):
    """
    Transforms galactic coordinates into celestial coordinates.

    Parameters
    ----------
    lon : array_like, float
        Galactic longitude, in degrees.
    b : array_like, float
        Galactic latitude, in degrees.
    dldt : array_like, float, optional
        Galactic longitude velocity, in mas/yr. The default is None
    dbdt : array_like, float, optional
        Galactic latitude velocity, in mas/yr. The default is None

    Returns
    -------
    a : array_like, float
        Right ascention in degrees.
    d : array_like, float
        Declination in degrees.
    dadt : array_like, float, optional
        Right ascention velocity (PMRA), in mas/yr.
    dddt : array_like, float, optional
        Declination velocity (PMDec), in mas/yr.

    """

    lon = lon * (np.pi / 180)
    b = b * (np.pi / 180)

    sind_NGP = np.sin(d_NGP)
    cosd_NGP = np.cos(d_NGP)
    sinb = np.sin(b)
    cosb = np.cos(b)
    cosdl = np.cos(l_NCP - lon)
    sindl = np.sin(l_NCP - lon)

    sind = sind_NGP * sinb + cosd_NGP * cosb * cosdl
    cosd_sinda = cosb * sindl
    cosd_cosda = cosd_NGP * sinb - sind_NGP * cosb * cosdl

    d = np.arcsin(sind)

    cosd = np.cos(d)

    sinda = cosd_sinda / cosd
    cosda = cosd_cosda / cosd

    if np.isscalar(sinda):
        if (sinda >= 0 and cosda > 0) or (sinda < 0 and cosda >= 0):
            a = a_NGP + np.arctan(cosd_sinda / cosd_cosda)
        elif (sinda > 0 and cosda <= 0) or (sinda <= 0 and cosda < 0):
            a = a_NGP + np.arctan(cosd_sinda / cosd_cosda) + np.pi
    else:
        a = np.zeros(len(sinda))
        for i in range(len(sinda)):
            if (sinda[i] >= 0 and cosda[i] > 0) or (sinda[i] < 0 and cosda[i] >= 0):
                a[i] = a_NGP + np.arctan(cosd_sinda[i] / cosd_cosda[i])
            elif (sinda[i] > 0 and cosda[i] <= 0) or (sinda[i] <= 0 and cosda[i] < 0):
                a[i] = a_NGP + np.arctan(cosd_sinda[i] / cosd_cosda[i]) + np.pi

    if np.isscalar(a):
        a = a % (2 * np.pi)
    else:
        for i in range(len(a)):
            a[i] = a[i] % (2 * np.pi)

    if dbdt is not None and dldt is not None:
        sind = np.sin(d)
        cosda = np.cos(a - a_NGP)

        dddt = (
            sind_NGP * cosb * dbdt
            - cosd_NGP * cosdl * sinb * dbdt
            + cosd_NGP * cosb * sindl * dldt
        ) / cosd

        dadt = -(
            cosd * (sinb * dbdt * sindl + cosb * cosdl * dldt)
            - cosb * sindl * sind * dddt
        ) / (cosd * cosd * cosda)

        dadt = dadt * cosd

        a = a * (180 / np.pi)
        d = d * (180 / np.pi)

        return a, d, dadt, dddt

    else:

        a = a * (180 / np.pi)
        d = d * (180 / np.pi)

        return a, d


def cart_to_radec(x, y, z, vx=None, vy=None, vz=None):
    """
    Transforms galactocentric coordinates into celestial coordinates.

    Parameters
    ----------
    x : array_like, float
        x-axis.
    y : array_like, float
        y-axis.
    z : array_like, float
        z-axis.
    vx : array_like, float, optional
        x-axis velocity. The default is None.
    vy : array_like, float, optional
        y-axis velocity. The default is None.
    vz : array_like, float, optional
        z-axis velocity. The default is None.

    Raises
    ------
    ValueError
        Velocity components are incomplete
        (number of velocity dimensions equal to 1 or 2).

    Returns
    -------
    a : array_like, float
        Right ascention in degrees.
    d : array_like, float
        Declination in degrees.
    r : array_float
        Distance of the source, in kpc.
    dadt : array_like, float, optional
        Right ascention velocity (PMRA), in mas/yr.
    dddt : array_like, float, optional
        Declination velocity (PMDec), in mas/yr.
    vr : array_like, float
        Line of sight velocity, in km/s.

    """
    if vx is None or vy is None or vz is None:

        onlypos = True

        if vx is not None:
            raise ValueError("Please provide other velocity components.")
        if vy is not None:
            raise ValueError("Please provide other velocity components.")
        if vz is not None:
            raise ValueError("Please provide other velocity components.")

        vx = 0
        vy = 0
        vz = 0

    else:

        onlypos = False

    lon, b, r, dldt, dbdt, vr = cart_to_lb(x, y, z, vx=vx, vy=vy, vz=vz)

    a, d, dadt, dddt = lb_to_radec(lon, b, dldt=dldt, dbdt=dbdt)

    if onlypos is True:
        return a, d, r
    else:
        return a, d, r, dadt, dddt, vr


def radec_to_cart(a, d, r, mua=None, mud=None, vr=None):
    """
    Transforms celestial coordinates into galactocentric coordinates.

    Parameters
    ----------

    a : array_like, float
        Right ascention in degrees.
    d : array_like, float
        Declination in degrees.
    r : array_float
        Distance of the source, in kpc.
    mua : array_like, float, optional
        Right ascention velocity (PMRA), in mas/yr.
        The default is None.
    mud : array_like, float, optional
        Declination velocity (PMDec), in mas/yr.
        The default is None.
    vr : array_like, float, optional
        Line of sight velocity, in km/s.
        The default is None.

    Raises
    ------
    ValueError
        Velocity components are incomplete
        (number of velocity dimensions equal to 1 or 2).

    Returns
    -------

    x : array_like, float
        x-axis.
    y : array_like, float
        y-axis.
    z : array_like, float
        z-axis.
    vx : array_like, float
        x-axis velocity.
    vy : array_like, float
        y-axis velocity.
    vz : array_like, float
        z-axis velocity.

    """

    masyr_to_kms = 4.7405 * r

    if mua is None or mud is None or vr is None:

        onlypos = True

        if mua is not None:
            raise ValueError("Please provide other velocity components.")
        if mud is not None:
            raise ValueError("Please provide other velocity components.")
        if vr is not None:
            raise ValueError("Please provide other velocity components.")

        mua = 0
        mud = 0
        vr = 0

    else:

        onlypos = False

    lon, b, dldt, dbdt = radec_to_lb(a, d, dadt=mua, dddt=mud)

    dldt = dldt * masyr_to_kms
    dbdt = dbdt * masyr_to_kms
    r = r * kpc_to_km

    phi = lon * (np.pi / 180)
    theta = np.pi * 0.5 - b * (np.pi / 180)

    x, y, z, vx, vy, vz = sph_to_cart(r, phi, theta, vr, dldt, -dbdt)

    x = -x / kpc_to_km + x_sun
    y = y / kpc_to_km - y_sun
    z = z / kpc_to_km - z_sun

    vx = -vx + vx_sun
    vy = vy + vy_sun
    vz = vz + vz_sun

    if onlypos is True:
        return x, y, z
    else:
        return x, y, z, vx, vy, vz


def cart_to_lb(x, y, z, vx=None, vy=None, vz=None):
    """
    Transforms galactocentric coordinates into galactic coordinates.

    Parameters
    ----------
    x : array_like, float
        x-axis.
    y : array_like, float
        y-axis.
    z : array_like, float
        z-axis.
    vx : array_like, float, optional
        x-axis velocity. The default is None.
    vy : array_like, float, optional
        y-axis velocity. The default is None.
    vz : array_like, float, optional
        z-axis velocity. The default is None.

    Raises
    ------
    ValueError
        Velocity components are incomplete
        (number of velocity dimensions equal to 1 or 2).

    Returns
    -------
    lon : array_like, float
        Galactic longitude, in degrees.
    b : array_like, float
        Galactic latitude, in degrees.
    r : array_float
        Distance of the source, in kpc.
    dldt : array_like, float, optional
        Galactic longitude velocity, in mas/yr.
    dbdt : array_like, float, optional
        Galactic latitude velocity, in mas/yr.
    vr : array_like, float
        Line of sight velocity, in km/s.

    """

    x = (-x + x_sun) * kpc_to_km
    y = (y + y_sun) * kpc_to_km
    z = (z + z_sun) * kpc_to_km

    if vx is None or vy is None or vz is None:

        onlypos = True

        if vx is not None:
            raise ValueError("Please provide other velocity components.")
        if vy is not None:
            raise ValueError("Please provide other velocity components.")
        if vz is not None:
            raise ValueError("Please provide other velocity components.")

        vx = 0
        vy = 0
        vz = 0

    else:

        onlypos = False

    vx = -vx + vx_sun
    vy = vy - vy_sun
    vz = vz - vz_sun

    r, phi, theta, vr, vphi, vtheta = cart_to_sph(x, y, z, vx, vy, vz)

    lon = phi * (180 / np.pi)
    b = (np.pi * 0.5 - theta) * (180 / np.pi)

    dldt = (vphi / r) * (3600 * 1000 * 180 / np.pi) * (3.154 * 10 ** 7)
    dbdt = (-vtheta / r) * (3600 * 1000 * 180 / np.pi) * (3.154 * 10 ** 7)
    r = r / kpc_to_km

    if onlypos is True:
        return lon, b, r
    else:
        return lon, b, r, dldt, dbdt, vr
